# -*- coding: utf-8 -*-
import ase
from ase.calculators import singlepoint
import pytest
import torch
import numpy as np

from milad import dat
from milad.models import neuralnetwork

torch.set_default_dtype(torch.float64)

# pylint: disable=invalid-name

SIGMA = 0.4
EPSILON = 4.


def lennard_jones(r):
    return EPSILON * ((SIGMA / r)**12 - (SIGMA / r)**6)


def lennard_jones_force(r):
    """Get magnitude of LJ force"""
    return -EPSILON * ((-12 * SIGMA**12 / r**13) + (6 * SIGMA**6 / r**7))


@pytest.fixture
def lj_training_data():
    training_size = 50
    r = np.linspace(0.3, 1., num=training_size)

    # Create training energies
    training_energies = np.array(list(map(lennard_jones, r)))
    # Create forces
    drs = np.outer(r, np.array([1.0, 0, 0]))  # Create a bunch of vectors that point along positive-x
    training_forces = np.array([force * dr for force, dr in zip(map(lennard_jones_force, r), drs)])

    training_data = []
    for dr, energy, force in zip(drs, training_energies, training_forces):
        atoms = ase.Atoms(positions=[[0., 0., 0.], dr], numbers=[6, 6])
        calc = singlepoint.SinglePointCalculator(atoms, energy=energy, forces=[-force, force])
        atoms.calc = calc

        training_data.append(atoms)

    return training_data


# pylint: disable=invalid-name, redefined-outer-name


def test_fitting_data(descriptor, training_data):
    fingerprint_set = dat.create_fingerprint_set(descriptor, training_data, get_derivatives=True)
    fitting_data = neuralnetwork.FittingData.from_fingerprint_set(fingerprint_set, requires_grad=True)

    num_systems = len(training_data)
    total_environments = sum([len(atoms) for atoms in training_data])

    assert len(fitting_data) == num_systems
    assert fitting_data.fingerprint_len == fingerprint_set.fingerprint_len
    assert len(fitting_data.fingerprints) == total_environments

    # Check that iterating system-wise works correctly
    idx = 0
    for system_idx, system in enumerate(fitting_data.iter_systemwise()):
        natoms = fitting_data.num_atoms[system_idx]
        assert fitting_data.total_energies[system_idx] == system.total_energy
        assert torch.all(fitting_data.derivatives[system_idx] == system.derivatives)

        assert torch.all(fitting_data.fingerprints[idx:idx + natoms] == system.fingerprints)
        assert torch.all(fitting_data.forces[idx:idx + natoms] == system.forces)
        assert torch.all(fitting_data.index[idx:idx + natoms] == system.index)

        idx += natoms

    total_atoms = 0
    for idx, atoms in enumerate(training_data):
        natoms = len(atoms)

        assert np.isclose(fitting_data.num_atoms[idx].item(), natoms)
        assert np.isclose(fitting_data.total_energies[idx].item(), atoms.get_potential_energy())
        assert np.isclose(fitting_data.get_normalised_energies()[idx].item(), atoms.get_potential_energy() / natoms)
        assert np.all(fitting_data.index[total_atoms:total_atoms + natoms].detach().numpy() == idx)

        total_atoms += natoms

    assert fitting_data.batch_split(len(training_data))[0] is fitting_data


def test_neural_network_basics(descriptor, training_data):
    nn = neuralnetwork.NeuralNetwork()
    fingerprint_set = dat.create_fingerprint_set(descriptor, training_data, get_derivatives=True)
    fitting_data = neuralnetwork.FittingData.from_fingerprint_set(fingerprint_set, requires_grad=True, device=nn.device)

    nn.init(fitting_data)

    predictions = nn.make_prediction(fitting_data, get_forces=True)
    assert torch.all(predictions.indices == fitting_data.index)
    assert torch.all(predictions.sizes == fitting_data.num_atoms)
    assert len(predictions.forces) == len(fitting_data.fingerprints), 'Expect one force per environment'

    nn.fit(fitting_data, max_epochs=1)


def test_predictions(descriptor, training_data):
    fingerprint_set = dat.create_fingerprint_set(descriptor, training_data, get_derivatives=True)
    fitting_data = neuralnetwork.FittingData.from_fingerprint_set(fingerprint_set, requires_grad=True)

    # Artificially create per-atom energies (just split the total energy evenly
    env_energies = torch.zeros((fitting_data.index.shape[0], 1), dtype=fitting_data.total_energies.dtype)
    total_atoms = 0
    for idx, energy in enumerate(fitting_data.get_normalised_energies()):
        natoms = fitting_data.num_atoms[idx]
        env_energies[total_atoms:total_atoms + natoms, 0] = energy
        total_atoms += natoms

    pred = neuralnetwork.Predictions(fitting_data.num_atoms, fitting_data.index, env_energies)
    predicted_energies = pred.get_normalised_energies()
    known_energies = fitting_data.get_normalised_energies()
    assert torch.allclose(predicted_energies, known_energies)

    loss_fn = neuralnetwork.LossFunction()
    loss = loss_fn.get_loss(pred, fitting_data)
    assert torch.isclose(loss.energy, torch.tensor(0.))
    assert torch.isclose(loss.total, torch.tensor(0.))


def test_nn_lennard_jones(descriptor, lj_training_data):
    """Check if we can successfully fit to the LJ potential"""
    nn = neuralnetwork.NeuralNetwork(hiddenlayers=(10, 10))

    fingerprint_set = dat.create_fingerprint_set(descriptor, lj_training_data, get_derivatives=True)
    _training, _loss_fn = nn.fit(fingerprint_set, max_epochs=50, batchsize=1000, learning_rate=1e-3)



# -*- coding: utf-8 -*-
import collections
import math
from typing import Union, Callable, Sequence, Optional, List, Tuple, Iterator

import ase

try:
    import functorch
except ImportError:
    functorch = None
import torch
import torch.nn.functional
from torch.autograd import functional

from milad import dat

__all__ = ('NeuralNetwork',)

# pylint: disable=no-member, not-callable


class Predictions:
    """Class representing predictions made by the the model"""

    def __init__(
        self, sizes: torch.Tensor, indices: torch.Tensor, local_energies: torch.Tensor, forces: torch.Tensor = None
    ):
        """Create an object to store prediction data from the neural network

        :param sizes: a tensor containing the number of atoms in each system.  The length of the tensor is the total
            number of systems the prediction is being made for.
        """
        if not sizes.sum() == len(local_energies):
            raise ValueError(
                f"The total number of atoms ({sizes.sum()}) doesn't match the number of "
                f'local energies ({len(local_energies)}).'
            )

        self.sizes = sizes
        self.indices = indices
        self.local_energies = local_energies
        self.forces = forces
        self._total_energies = None

    @property
    def total_energies(self) -> torch.Tensor:
        """Get total energies as a single tensor that preserves gradients"""
        if self._total_energies is None:
            # Lazily calculate
            device = self.local_energies.device

            # Combine them all into a tensor
            self._total_energies = torch.zeros(len(self.sizes), device=device, dtype=self.local_energies.dtype)
            self._total_energies.scatter_add_(0, self.indices, self.local_energies[:, 0])

        return self._total_energies

    def get_normalised_energies(self) -> torch.Tensor:
        """Get the energy/atom for each system"""
        return self.total_energies / self.sizes


SystemFittingData = collections.namedtuple(
    'SystemFittingData', 'index num_atoms total_energy forces fingerprints derivatives'
)


class FittingData:
    """Data used during a fitting procedure"""

    @classmethod
    def from_fingerprint_set(cls, fingerprint_set: dat.FingerprintSet, device=None, dtype=None, requires_grad=False):
        tensor_kwargs = dict(dtype=dtype, device=device, requires_grad=requires_grad)
        fingerprints = torch.cat(tuple(torch.tensor(fps, **tensor_kwargs) for fps in fingerprint_set.fingerprints))

        forces = None
        if fingerprint_set.has_all_forces():
            forces = torch.cat(
                tuple(
                    torch.tensor(forces, **tensor_kwargs) if forces is not None else None
                    for forces in fingerprint_set.get_forces()
                )
            )

        derivatives = None
        if fingerprint_set.has_all_derivatives():
            derivatives = tuple(
                torch.tensor(derivatives, **tensor_kwargs) if derivatives is not None else None
                for derivatives in fingerprint_set.fingerprint_derivatives
            )

        return FittingData(
            torch.tensor(fingerprint_set.sizes, dtype=torch.int64, device=device),
            fingerprints,
            total_energies=torch.tensor(fingerprint_set.get_potential_energies(normalise=False), **tensor_kwargs),
            forces=forces,
            derivatives=derivatives
        )

    def __init__(
        self,
        sizes: torch.Tensor,
        fingerprints: torch.Tensor,
        total_energies: torch.Tensor,
        forces: Optional[torch.Tensor] = None,
        derivatives: Tuple[Optional[torch.Tensor]] = None
    ):
        """
        Construct a set of fitting data.

        :param sizes: a tensor containing the number of atoms in each system.  The length of the tensor should be equal
            to that of total_energies
        :param fingerprints: a tuple containing a set of fingerprints for each atom in the structure
        :param total_energies: the corresponding energy for each structure
        :param forces: the (optional) forces for each atom in the structure
        """
        if sizes.sum() != len(fingerprints):
            raise ValueError(
                f"The total number of atoms ({sizes.sum()}) doesn't match the number of "
                f'fingerprints ({len(fingerprints)}).'
            )
        if len(sizes) != len(total_energies):
            raise ValueError(
                f"The number of total energies ({len(total_energies)}) doesn't match the number of "
                f'systems (sizes) ({len(sizes)}).'
            )
        if derivatives and len(derivatives) != len(sizes):
            raise ValueError("Number of system-wise derivatives doesn't match the number of sizes")

        self._num_atoms = sizes  # Per-system
        self._total_energies = total_energies  # Per-system
        self._derivatives = derivatives  # Per system

        self._fingerprints = fingerprints  # Per-environment
        self._forces = forces  # Per-environment
        self._index = _create_index_from_sizes(self._num_atoms)  # Per-environment

    def __getitem__(self, item) -> 'FittingData':
        if isinstance(item, slice):
            atomic_slice = self._get_per_atom_slice(item)

            sizes = self._num_atoms[item]
            total_energies = self._total_energies[item]
            fingerprints = self._fingerprints[atomic_slice]
            forces = self._forces[atomic_slice] if self._forces is not None else None
            derivatives = self._derivatives[item] if self._derivatives is not None else None

            return FittingData(sizes, fingerprints, total_energies, forces, derivatives)

        # Assume item is a fixed index
        return self.__getitem__(slice(item, item + 1))

    def __len__(self) -> int:
        """Returns the total number of systems contained in this fitting data"""
        return len(self._num_atoms)

    def _get_per_atom_slice(self, item: slice) -> slice:
        if item.step is not None and item.step != 1:
            raise ValueError(f'Non-contigious slices are unsupported for now, got step {item.step}')

        start = self._num_atoms[:item.start].sum().cpu().item()
        stop = self._num_atoms[:item.stop].sum().cpu().item()
        return slice(start, stop)

    def iter_systemwise(self) -> Iterator[SystemFittingData]:
        env_idx = 0
        for system_idx, natoms in enumerate(self._num_atoms):
            env_slice = slice(env_idx, env_idx + natoms)
            yield SystemFittingData(
                index=system_idx,
                num_atoms=natoms,
                total_energy=self._total_energies[system_idx],
                derivatives=self._derivatives[system_idx],
                fingerprints=self._fingerprints[env_slice],
                forces=self._forces[env_slice],
            )
            env_idx += natoms

    @property
    def fingerprint_len(self) -> int:
        """Get the length of the fingerprint vector"""
        return self._fingerprints.shape[1]

    @property
    def fingerprints(self) -> torch.Tensor:
        """Fingerprints tensors.  One entry per atomic environment."""
        return self._fingerprints

    @property
    def total_energies(self) -> torch.Tensor:
        """Get the known energies"""
        return self._total_energies

    @property
    def forces(self) -> Optional[torch.Tensor]:
        """Get the known forces if we have them"""
        return self._forces

    @property
    def derivatives(self) -> Tuple[Optional[torch.Tensor]]:
        """Get the derivatives for each environment in the structure.  This returns a tuple with one entry per system.

        Each entry is a tensor in the format [central atom idx, neighbour idx, 3 (x, y, z), fingerprint length]"""
        return self._derivatives

    @property
    def num_atoms(self) -> torch.Tensor:
        """Get the number of atoms in each structure"""
        return self._num_atoms

    @property
    def index(self) -> torch.Tensor:
        """Get a tensor of structure indices that each local environment datum belongs to
        (e.g. force, derivatives, etc)"""
        return self._index

    def get_normalised_energies(self) -> torch.Tensor:
        return self._total_energies / self._num_atoms

    def batch_split(self, batchsize: int) -> List['FittingData']:
        total_samples = len(self)
        num_batches = int(math.ceil(total_samples / batchsize))
        if num_batches == 1:
            return [self]

        batches = []
        for batch_num in range(num_batches):
            start_idx = batch_num * batchsize
            end_idx = min(start_idx + batchsize, total_samples)
            batches.append(self[start_idx:end_idx])

        return batches


class LossFunction:
    Loss = collections.namedtuple('Result', 'energy force total')

    def __init__(self, energy_coeff=1., force_coeff=0.1):
        self.energy_coeff = energy_coeff
        self.force_coeff = force_coeff

    def get_loss(self, predictions: Predictions, fitting_data: FittingData):
        total_loss = 0.
        energy_loss = torch.nn.functional.mse_loss(
            predictions.get_normalised_energies(), fitting_data.get_normalised_energies()
        )
        total_loss += self.energy_coeff * energy_loss

        force_loss = None
        if fitting_data.forces is not None and predictions.forces is not None and self.force_coeff != 0.:
            force_loss = torch.nn.functional.mse_loss(predictions.forces, fitting_data.forces) / 3.
            total_loss += self.force_coeff * force_loss
            total_loss *= 0.5

        return LossFunction.Loss(energy_loss, force_loss, total_loss)


class Range:
    """A range object representing a minimum and maximum value"""

    def __init__(self, initial=(0., 1.), device=None, dtype=None):
        initial = initial or (0., 1.)
        self._range = torch.tensor(initial, device=device, dtype=dtype)
        self._using_defaults = True

    def __str__(self):
        return '[{}, {}]'.format(self.min.item(), self.max.item())

    @property
    def range(self):
        return self._range

    @property
    def min(self):
        return self._range[0]

    @property
    def max(self):
        return self._range[1]

    @property
    def span(self):
        return self.max - self.min

    def expand(self, values: torch.Tensor):
        """Expand the range to encompass the passed values"""
        if self._using_defaults:
            self._range[0] = values.min()
            self._range[1] = values.max()
            self._using_defaults = False
        else:
            vals_min = values.min()
            vals_max = values.max()
            self._range[0] = self.min if self.min < vals_min else vals_min
            self._range[1] = self.max if self.max > vals_max else vals_max


class GlobalScaler(torch.nn.Module):
    """Scale data to be in the range in a particular range.  Handles scaling on both input and output sides
    so if you can set the range of input and outputs separately."""

    def __init__(self, in_range=(0., 1.), out_range=(0., 1.), device=None, dtype=None):
        super().__init__()
        self._in_range = Range(in_range, device=device, dtype=dtype)  # Input range
        self._out_range = Range(out_range, device=device, dtype=dtype)  # Output range

    def __str__(self):
        return '{} -> {}'.format(self._in_range, self._out_range)

    @property
    def input(self) -> Range:
        return self._in_range

    @property
    def output(self) -> Range:
        return self._out_range

    def scale(self, x: torch.Tensor):  # pylint: disable=invalid-name
        return self.output.span * (x - self.input.min) / self.input.span + self.output.min

    def unscale(self, y: torch.Tensor):  # pylint: disable=invalid-name
        return (y - self.output.min) * self.input.span / self.output.span + self.input.min

    def forward(self, x: torch.Tensor):  # pylint: disable=invalid-name
        return self.scale(x)


class ComponentWiseInputScaler(torch.nn.Module):

    def __init__(self, inputs: torch.Tensor):
        super().__init__()
        xmin = torch.min(inputs, dim=0)[0]
        xmax = torch.max(inputs, dim=0)[0]
        delta = xmax - xmin

        self._scale = 2 / delta
        self._offset = (2 * xmin + delta) / delta

        zeros = delta.isclose(torch.tensor(0., dtype=delta.dtype))
        self._scale[zeros] = 1.
        self._offset[zeros] = 0.

    def forward(self, inp: torch.Tensor) -> torch.Tensor:
        return inp * self._scale - self._offset


class NeuralNetwork:
    """Neural network empirical potential model"""

    def __init__(
        self,
        hiddenlayers=(32, 16, 8),
        activations: Union[str, Callable] = 'Tanh',
        loss_function=None,
        device=None,
        dtype=torch.float64,
        bias=False,
        vectorise_jacobian=False,
    ):
        """

        :param hiddenlayers: a tuple of the hidden layers to use e.g. (32, 16, 8)
        :param activations:
        :param loss_function:
        :param device:
        :param dtype:
        :param bias:
        :param vectorise_jacobian: if True, will calculate the NN Jacobian for a batch of fingerprints all in one go.
            This is very memory consuming at the moment and may exceed the amount of memory you have available.
        """
        # Set up the hidden layers
        self._hidden = hiddenlayers

        # Set the activation function
        if isinstance(activations, str):
            self._activations = getattr(torch.nn, activations)()
        elif isinstance(activations, Callable):  # pylint: disable=isinstance-second-argument-not-valid-type
            self._activations = activations
        else:
            raise TypeError('Expecting str or Callable, got {}'.format(activations.__class__.__name__))

        if device is None:
            self._device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        else:
            self._device = device
        self._dtype = dtype
        self._bias = bias
        self.vectorise_jacobian: bool = vectorise_jacobian

        self._fingerprint_scaler = GlobalScaler(out_range=(-1., 1.), device=self._device, dtype=self._dtype)
        self._energy_scaler = GlobalScaler(in_range=(-1., 1.), device=self._device, dtype=self._dtype)
        self.loss_function = loss_function or LossFunction()

        self._network = None

    def init(self, training_data: FittingData):
        """Reinitialise the model.  This will throw away anything learned so far"""
        self._create_network(training_data)

    @property
    def device(self):
        return self._device

    def fit(
        self,
        training_set: Union[Sequence[ase.Atoms], dat.FingerprintSet],
        max_epochs=200,
        progress_callback: Callable = None,
        learning_rate=5e-4,
        batchsize=16,
    ) -> Tuple[FittingData, LossFunction]:
        if isinstance(training_set, FittingData):
            training_data = training_set
        else:
            training_data = self.create_fitting_data(training_set, requires_grad=True)

        if self._network is None:
            self._create_network(training_data)

        def stopping_function(epoch, _training, _loss):
            if epoch >= max_epochs:
                return True

            return False  # Don't stop

        optimiser = torch.optim.AdamW(self._network.parameters(), lr=learning_rate)
        # optimiser = torch.optim.SGD(self._network.parameters(), lr=learning_rate, momentum=0.9)
        # optimiser = torch.optim.LBFGS(self._network.parameters(), lr=learning_rate)
        loss = self._train(
            training_data,
            stopping_function,
            optimiser,
            batchsize,
            progress_callback=progress_callback,
        )
        return training_data, loss

    def create_fitting_data(self, fingerprint_set: dat.FingerprintSet, requires_grad=False):
        return FittingData.from_fingerprint_set(
            fingerprint_set,
            device=self._device,
            dtype=self._dtype,
            requires_grad=requires_grad,
        )

    def _train(
        self,
        training: FittingData,
        should_stop: Callable,
        optimiser,
        batchsize: int,
        progress_callback: Callable = None,
    ) -> LossFunction.Loss:
        # Let's break up the training data into batches
        batches = training.batch_split(batchsize)

        epoch = 0
        loss_result = None
        get_forces = self.loss_function.force_coeff != 0.
        while True:
            for batch in batches:
                optimiser.zero_grad()

                predictions = self.make_prediction(batch, get_forces=get_forces, create_graph=True)
                loss_result = self.loss_function.get_loss(predictions, batch)

                # Ask the optimiser to make a step
                loss_result.total.backward(retain_graph=True)

                optimiser.step()

            # Use no_grad to reduce memory footprint
            with torch.no_grad():
                # Calculate loss for the entire training set.  Only necessary if there is more than one batch
                if len(batches) > 1:
                    predictions = self.make_prediction(training, get_forces=get_forces)
                    loss_result = self.loss_function.get_loss(predictions, training)

                if progress_callback is not None:
                    progress_callback(self, epoch, training, loss_result)

                if should_stop(epoch, training, loss_result):
                    break

            epoch += 1

        return loss_result

    def loss(self, fitting_data: FittingData, get_forces=None) -> LossFunction.Loss:
        get_forces = get_forces if get_forces is not None else self.loss_function.force_coeff != 0.
        predictions = self.make_prediction(fitting_data, get_forces=get_forces)
        return self.loss_function.get_loss(predictions, fitting_data)

    def make_prediction(self, fitting_data: FittingData, get_forces=False, create_graph=False) -> Predictions:
        # pylint: disable=too-many-locals,too-many-nested-blocks
        # Join all fingerprints as this is much faster going through the network
        local_energies = self._network(fitting_data.fingerprints)

        forces = None
        if get_forces:
            # Need to calculate the forces using the chain rule with the fingerprints derivative and the
            # neural network derivative
            if self.vectorise_jacobian:
                if functorch is not None:
                    # VMAP version, This came from a discussion with Richard Zou about.  See:
                    # https://github.com/pytorch/pytorch/issues/42368

                    jacs = functorch.vmap(functorch.jacrev(self._network))(fitting_data.fingerprints)
                    forces = -torch.matmul(jacs, fitting_data.derivatives)[:, 0, :]
                else:
                    forces = []

                    jac = functional.jacobian(
                        self._network, fitting_data.fingerprints, create_graph=create_graph, vectorize=True
                    )
                    jacs = [jac[i, :, i, :] for i in range(len(fitting_data.fingerprints))]

                    for _fingerprint, derivatives, network_deriv in zip(
                        fitting_data.fingerprints, fitting_data.derivatives, jacs
                    ):
                        # Get the derivative of the energy wrt to input vector
                        predicted_force = -torch.matmul(network_deriv, derivatives)
                        forces.append(predicted_force)

                    forces = torch.cat(forces)
            else:
                # Fallback: lower memory, but slow, version
                forces = []

                for system in fitting_data.iter_systemwise():
                    natoms = system.num_atoms
                    denergy_dphi = tuple(
                        functional.jacobian(self._network, fingerprint, create_graph=create_graph, strict=True)[0]
                        for fingerprint in system.fingerprints
                    )
                    # Calculate the force on each atom
                    for i in range(natoms):
                        total_force = torch.zeros(3, device=self.device)
                        for j in range(natoms):
                            if not torch.all(system.derivatives[j, i, :, :] == 0):
                                total_force += -torch.matmul(system.derivatives[j, i, :, :], denergy_dphi[j])
                        forces.append(total_force)

                forces = torch.stack(forces)

        return Predictions(fitting_data.num_atoms, fitting_data.index, local_energies=local_energies, forces=forces)

    def _create_network(self, training_data: FittingData):
        # Create the network
        sequence = [ComponentWiseInputScaler(training_data.fingerprints)]

        prev_size = training_data.fingerprint_len
        # Create the fully connected (hidden) layers
        for size in self._hidden:
            sequence.append(torch.nn.Linear(prev_size, size, bias=self._bias))
            sequence.append(self._activations)
            prev_size = size
        # Add the output layer
        sequence.extend((torch.nn.Linear(prev_size, 1, bias=self._bias), self._energy_scaler))

        # Scale the energy outputs to match the ranges
        self._energy_scaler.output.expand(training_data.get_normalised_energies())

        self._network = torch.nn.Sequential(*sequence)
        self._network.to(self._device, self._dtype)

    def tensor(self, *args, requires_grad=False) -> torch.Tensor:
        return torch.tensor(*args, device=self._device, dtype=self._dtype, requires_grad=requires_grad)


def _create_index_from_sizes(sizes: torch.Tensor) -> torch.Tensor:
    device = sizes.device
    index = torch.empty(sizes.sum(), dtype=torch.int64, device=device)
    idx = 0
    for i, num_atoms in enumerate(sizes):
        index[idx:idx + num_atoms] = i
        idx += num_atoms

    return index



# -*- coding: utf-8 -*-
from typing import Tuple

import numpy as np
from scipy.spatial import transform

from . import mathutil


def random_points_in_sphere(num: int, radius=1., centre=False, minsep=None) -> np.array:
    """Generate num points within a sphere of the given radius"""
    points = np.empty((num, 3))
    r_sq = radius * radius

    for idx in range(num):
        vec = 2 * radius * (np.random.rand(3) - 0.5)
        while np.dot(vec, vec) >= r_sq:
            vec = 2 * radius * (np.random.rand(3) - 0.5)
        points[idx] = vec

    if minsep is not None:
        pass

    if centre:
        points -= centeroid(points)
        # Make sure that there aren't any outside of the radius now
        radius_sq = (points**2).sum(1).max()
        if radius_sq > r_sq:
            points *= r_sq / radius_sq

    return points


def centeroid(pts: np.ndarray) -> np.ndarray:
    length = len(pts)
    centroid = np.empty((length, 3))
    centroid[:, 0] = np.sum(pts[:, 0]) / length
    centroid[:, 1] = np.sum(pts[:, 1]) / length
    centroid[:, 2] = np.sum(pts[:, 2]) / length
    return centroid


def chiral_tetrahedra() -> Tuple[np.ndarray, np.ndarray]:
    """Create chiral structures that cannot be distinguished by bispectrum from
    https://link.aps.org/doi/10.1103/PhysRevLett.125.166001
    """
    # pylint: disable=invalid-name
    a = np.radians(0)
    b = np.radians(-87)
    c = np.radians(176)

    # First group of points
    b_i = np.zeros((3, 3))
    r = 0.3
    b_i[0, :] = mathutil.sph2cart([r, a, 0.])
    b_i[1, :] = mathutil.sph2cart([r, b, 0.])
    b_i[2, :] = mathutil.sph2cart([r, c, 0.])

    # Second group of points
    rot = transform.Rotation.from_euler('y', 86, degrees=True)
    b_i_prime = rot.apply(b_i)

    pts_dist = 0.4
    b_i[:, 1] += pts_dist
    b_i_prime[:, 1] -= pts_dist

    pt_dist = 0.8

    plus = np.concatenate((np.array([[0, 0, 0], [0, +pt_dist, 0.]]), b_i, b_i_prime))
    minus = np.concatenate((np.array([[0, 0, 0], [0, -pt_dist, 0.]]), b_i, b_i_prime))

    return minus, plus



# -*- coding: utf-8 -*-
import random

import numpy as np
import pytest
from scipy import spatial
import sympy

from milad import atomic
from milad import optimisers
from milad import testing


def test_map_numbers():
    """Test that the mapping atomic numbers function works correctly"""
    num_atoms = 6
    system = atomic.AtomsCollection(
        num_atoms, positions=np.random.rand(num_atoms, 3), numbers=np.array([1, 2, 5, 2, 3, 8])
    )
    numbers = set(system.numbers)

    # Map the five species into the range 2 -> 3
    mapper = atomic.MapNumbers(numbers, map_to=(2., 3.))
    mapped = mapper(system)
    assert mapped.numbers[0] == pytest.approx(2.1)
    assert mapped.numbers[1] == pytest.approx(2.3)
    assert mapped.numbers[2] == pytest.approx(2.7)
    assert mapped.numbers[3] == pytest.approx(2.3)
    assert mapped.numbers[4] == pytest.approx(2.5)
    assert mapped.numbers[5] == pytest.approx(2.9)

    # Now test inversion
    inverted = mapper.inverse(mapped)
    diffs = inverted.numbers - system.numbers
    assert diffs.max() == pytest.approx(0.)

    testing.test_function(mapper, system.copy())

    # Now check that we can also map all species down to a single number (i.e. discard species info)
    mapper = atomic.MapNumbers(numbers, map_to=3.4)
    mapped = mapper(system)
    assert np.all(mapped.numbers == 3.4)


def test_separation_force():
    """Test that the separation force works correctly"""
    separation_force = atomic.SeparationForce()
    separation = 0.5
    atoms = atomic.AtomsCollection(2, positions=[[-0.5 * separation, 0, 0], [0.5 * separation, 0, 0]], numbers=1.)

    energy, jacobian = separation_force(atoms, jacobian=True)
    assert energy == separation_force.energy(separation)
    assert energy > 0., 'The energy should always be positive if the separation less than the cutoff'

    force = separation_force.force(separation)
    assert force > 0., 'The force should act to separate the two atoms'

    assert np.all(jacobian[0, atoms.linear_pos_idx(0)] == np.array([force, 0., 0.]))
    assert np.all(jacobian[0, atoms.linear_pos_idx(1)] == np.array([-force, 0., 0.]))

    # Check that the cutoff works
    atoms.positions[1, 0] = 1.
    energy, jacobian = separation_force(atoms, jacobian=True)
    assert energy == 0.0
    assert np.all(jacobian == 0.)


def test_separation_force_symbolic():
    # pylint: disable=invalid-name
    rcut = 1.5
    force = atomic.SeparationForce(cutoff=rcut)
    # Manually set the cutoff to None because otherwise the symbolic derivative check can't decide if
    # it is or isn't within the cutoff
    force._cutoff = None  # pylint: disable=protected-access

    r = sympy.Symbol('r', real=True)
    energy = force.energy(r)
    derivative = sympy.diff(energy, r)
    assert testing.sympy_equal(force.force(r), -derivative)

    energy_at_rcut = force.energy(rcut)
    assert energy_at_rcut == 0.

    force_at_rcut = force.force(rcut)
    assert force_at_rcut == 0.


def test_separation_force_optimiser():
    """Test using the separation force to actually separate some atoms using gradient descent"""
    cutoff = 1.
    force = atomic.SeparationForce(epsilon=0.1, sigma=0.5, cutoff=cutoff)
    atoms = atomic.random_atom_collection_in_sphere(10, radius=0.3)

    res = optimisers.LeastSquaresOptimiser().optimise(force, initial=atoms, verbose=True)

    assert res.success
    distances = spatial.distance.pdist(res.value.positions)
    # Need to leave a bit of a tolerance on the cutoff as the condition need not be _exactly_ satisfied
    assert np.all(distances > (cutoff - 1e-1))


def test_scale_positions():
    atoms = atomic.random_atom_collection_in_sphere(10, radius=5., centre=True)
    testing.test_function(atomic.ScalePositions(0.5), atoms, check_jacobian=True)


def test_feature_mapper():
    atoms = atomic.random_atom_collection_in_sphere(10, numbers=tuple(random.randint(0, 100) for _ in range(10)))
    testing.test_function(atomic.FeatureMapper(map_species_to='WEIGHT'), atoms, check_jacobian=True)



# -*- coding: utf-8 -*-
import numpy as np

import milad
from milad import atomic
from milad import functions
from milad import invariants
from milad import testing

# pylint: disable=invalid-name


def test_descriptor_derivatives():
    invs = invariants.read(invariants.COMPLEX_INVARIANTS, max_order=3)

    descriptor = milad.descriptor(
        features=dict(type=functions.WeightedDelta, map_species_to=None), cutoff=5., apply_cutoff=False, invs=invs
    )

    structure = atomic.random_atom_collection_in_sphere(1, 5.)
    testing.test_function(descriptor, structure)


def test_descriptor_smoothness(descriptor):
    last_max = 0.
    for x in np.linspace(1.01, 0.9, 12):
        atoms = milad.AtomsCollection(1, positions=[x, 0., 0.], numbers=1)
        fingerprint = descriptor(atoms)

        fingerprint_max = np.abs(fingerprint).max()

        if x >= 1.:
            assert fingerprint_max == 0.

        assert fingerprint_max >= last_max

        last_max = fingerprint_max



# -*- coding: utf-8 -*-
import numpy as np

from milad import mathutil

# pylint: disable=invalid-name


def test_spherical_cart():
    """Test conversion to and from cartesian/spherical coords works correctly"""
    x = np.random.rand(3)
    r, theta, phi = mathutil.cart2sph(x)
    xprime = np.array(mathutil.sph2cart([r, theta, phi]))
    assert np.allclose(x, xprime)



# -*- coding: utf-8 -*-
from typing import Iterator

import numpy as np

__all__ = ('calculate_all_pair_distances',)


def calculate_all_pair_distances(vectors, sort_result=True):
    """Calculate all pair distances between the given vectors"""
    num = len(vectors)
    lengths = []
    for i in range(num - 1):
        for j in range(i + 1, num):
            dr = vectors[i] - vectors[j]  # pylint: disable=invalid-name
            lengths.append(np.linalg.norm(dr))

    if sort_result:
        lengths.sort()
    return lengths


def even(val: int) -> bool:
    """Test if an integer is event.  Returns True if so."""
    return (val % 2) == 0


def odd(val: int) -> bool:
    """Test if an integer is odd.  Returns True if so."""
    return not even(val)


def inclusive(*args) -> Iterator[int]:
    """Like range() but inclusive of upper bound and automatically does iteration of ranges with a
    negative step e.g. 0, -4 will produce a range containing 0, -1, -2, -3, -4"""
    if len(args) not in (1, 2, 3):
        raise ValueError('Takes one or two args, got: {}'.format(args))

    if len(args) == 3:
        # Assume form is start, stop, step
        start, stop, step = args
    else:
        if len(args) == 1:
            start = 0
            stop = args[0]
        else:
            start = args[0]
            stop = args[1]

        step = 1 if start <= stop else -1

    sign = 1 if step > 0 else -1
    idx = start
    while sign * (stop - idx) >= 0:
        yield idx
        idx += step


def outer_product(*array) -> np.array:
    if not array:
        raise ValueError('No arrays supplied')

    product = array[0]
    for entry in array[1:]:
        product = np.tensordot(product, entry, axes=0)

    return product



# -*- coding: utf-8 -*-
import numpy as np

from milad import utils
from . import interfaces

__all__ = ('PowerSpectrum',)


class PowerSpectrum(interfaces.Invariants):
    """Calculate the power spectrum given a set of spherical harmonic based expansion coefficients"""
    supports_jacobian = False

    def __init__(self, mix_radials=True, radials_first=False):
        super().__init__()
        self._mix_radials = mix_radials
        self._radials_first = radials_first

    def evaluate(
        # pylint: disable=unused-argument
        self,
        state,
        *,
        get_jacobian=False
    ):
        # pylint: disable=invalid-name

        indices = state.indices
        invariants = []

        if self._radials_first:
            nmax = indices.n.max
            # Loop over radial functions first
            for n1 in utils.inclusive(nmax):
                n2_idx = utils.inclusive(n1, nmax, 2) if self._mix_radials else [n1]
                for n2 in n2_idx:
                    for l in indices.iter_l(min(n1, n2)):
                        invariants.append(
                            np.vdot(state.array[n1, l, :].compressed(), state.array[n2, l, :].compressed()).real
                        )
        else:
            lmax = indices.l.max

            # Loops over angular frequencies first
            for l in utils.inclusive(lmax):
                for n1 in indices.iter_n(l):
                    n2_idx = indices.iter_n(l, (n1, None)) if self._mix_radials else [n1]
                    for n2 in n2_idx:
                        invariants.append(
                            np.vdot(state.array[n1, l, :].compressed(), state.array[n2, l, :].compressed()).real
                        )

        return np.array(invariants)



# -*- coding: utf-8 -*-
"""Module that is concerned with the calculation of moment invariants"""
import collections
import functools
import operator
import pathlib
from typing import Sequence, Union, List, Set, Tuple, Iterator, Callable

import numpy as np

from milad import base_moments
from milad import functions
from milad import geometric
from milad import polynomials

__all__ = ('MomentInvariant', 'read_invariants', 'RES_DIR', 'COMPLEX_INVARIANTS', 'GEOMETRIC_INVARIANTS', \
           'MomentInvariants', 'read', 'InvariantBuilder')

# The resources directory
RES_DIR = pathlib.Path(__file__).parent / 'res'
GEOMETRIC_INVARIANTS = RES_DIR / 'rot3dinvs8mat.txt'
COMPLEX_INVARIANTS = RES_DIR / 'cmfs7indep_0.txt'
COMPLEX_INVARIANTS_ORIG = RES_DIR / 'cmfs7indep_0.orig.txt'
# A convenience map to make it easier to load the default invariants
INVS_MAP = {'geometric': GEOMETRIC_INVARIANTS, 'complex': COMPLEX_INVARIANTS, 'complex-orig': COMPLEX_INVARIANTS_ORIG}


def prod(iterable):
    return functools.reduce(operator.mul, iterable, 1)


class MomentInvariant(polynomials.HomogenousPolynomial):
    """Class storing moment invariants.

    The invariants consist of a sum of terms where each term has a prefactor multiplied by a product
    of moments labelled by three indices e.g. p * c_20^0 * c_20^0
    """

    def __init__(self, weight: int, *term, constant=0, conjugate_values=False):
        factors, arr = zip(*term)
        super().__init__(weight, factors, arr, constant=constant, conjugate_values=conjugate_values)
        self._weight = weight
        self._max_order = -1
        # self._constant = constant

        self._indarray = None  # The index array
        self._norm_power = 1

        # self._build()

    @property
    def weight(self) -> int:
        """The number of terms in each product of the invariant.  This gives the units"""
        return self._weight

    @property
    def max_order(self) -> int:
        return self._terms.max()

    @property
    def terms_array(self) -> np.ndarray:
        return self.terms

    def apply(self, raw_moments: base_moments.Moments, normalise=False) -> float:
        """Compute this invariant from the given moments optionally normalising"""
        total = self.evaluate(raw_moments)
        if normalise:
            return total / raw_moments[0, 0, 0]**self._norm_power

        return total

    def __call__(self, *args, **kwargs):
        return self.apply(*args, **kwargs)

    def evaluate(self, values: Union[np.ndarray, base_moments.Moments]):
        if not isinstance(values, np.ndarray):
            # If we can get a matrix we can still use the fast (numpy) method
            try:
                values = values.array
            except AttributeError:
                pass

        return super().evaluate(values)


class InvariantBuilder:
    """Tools that can be used to build an invariant term by term"""

    def __init__(self, weight: int, reduce=True):
        self._weight = weight
        self._terms = []
        self._constant = 0
        self._reduce = reduce

    @property
    def constant(self):
        return self._constant

    @constant.setter
    def constant(self, new_value):
        self._constant = new_value

    def add_term(self, prefactor, indices: Sequence[Tuple]):
        """
        :param prefactor: the prefactor for this term
        :param indices: the indices of the moments multiplied together in this term
        """
        if not indices:
            # If there are no indices supplied then it's just a constant
            self._constant += prefactor
            return

        if not all(len(entry) == 3 for entry in indices):
            raise ValueError('There have to be three indices per entry, got: {}'.format(indices))
        self._terms.append((prefactor, tuple(indices)))

    def build(self) -> MomentInvariant:
        """Returns the invariant from the current set of terms"""
        terms = self._terms
        if self._reduce:
            terms_dict = collections.defaultdict(set)
            for i, (_, product) in enumerate(self._terms):
                terms_dict[tuple(sorted(product))].add(i)

            if len(terms) != len(terms_dict):
                # Now build up the new terms
                terms = list()
                for idx_set in terms_dict.values():
                    prefactor = sum(self._terms[idx][0] for idx in idx_set)
                    terms.append((prefactor, self._terms[tuple(idx_set)[0]][1]))

        inv = MomentInvariant(self._weight, *terms, constant=self._constant)
        return inv


class MomentInvariants(functions.Function):
    """A function that takes moments as input and produces rotation invariants using polynomials thereof"""
    input_type = base_moments.Moments
    output_type = np.ndarray
    supports_jacobian = True
    dtype = None

    def __init__(self, *invariant: MomentInvariant, are_real=True):
        super().__init__()
        for entry in invariant:
            if not isinstance(entry, MomentInvariant):
                raise TypeError(f'Expected MomentInvariant, got {entry.__class__.__name__}')

        self._invariants: List[MomentInvariant] = list(invariant)
        self._max_order = -1
        self._real = are_real

    def __len__(self) -> int:
        """Get the total number of invariants"""
        return len(self._invariants)

    def __iter__(self) -> Iterator[MomentInvariant]:
        """Iterate the invariants"""
        return self._invariants.__iter__()

    def __getitem__(self, item) -> Union['MomentInvariants', MomentInvariant]:
        """Get particular invariant(s)"""
        if isinstance(item, slice):
            return MomentInvariants(*self._invariants[item])
        if isinstance(item, tuple):
            if len(item) == 1:
                return MomentInvariants(self._invariants[item[0]])
            return MomentInvariants(*operator.itemgetter(*item)(self._invariants))

        return self._invariants[item]

    def filter(self, func: Callable) -> 'MomentInvariants':
        """Return moment invariants for which the passed callable returns true"""
        return MomentInvariants(*filter(func, self._invariants), are_real=self._real)

    def find(self, func: Callable) -> Tuple[int]:
        """Find the indices of invariants where fun(inv) returns True"""
        return tuple(i for i, inv in enumerate(self._invariants) if func(inv))

    @property
    def max_order(self) -> int:
        """Get the maximum order of all the invariants"""
        if self._max_order == -1:
            max_order = 0
            for inv in self._invariants:
                max_order = max(max_order, inv.max_order)
            self._max_order = max_order

        return self._max_order

    @property
    def variables(self) -> Set[Tuple]:
        """Return a set of all the the indices used by these invariants"""
        indices = set()
        for inv in self._invariants:
            indices.update(inv.variables)
        return indices

    def output_length(self, _in_state: functions.State) -> int:  # pylint: disable=unused-argument
        return len(self._invariants)

    def apply(self, moms: np.array, normalise=False, results=None) -> list:
        """Calculate the invariants from the given moments"""
        return apply_invariants(self._invariants, moms, normalise=normalise, results=results)

    def append(self, invariant: MomentInvariant):
        """Add an invariant"""
        self._invariants.append(invariant)
        self._max_order = max(self._max_order, invariant.max_order)

    def evaluate(self, moments: base_moments.Moments, *, get_jacobian=False) -> np.ndarray:  # pylint: disable=arguments-differ
        vector = np.empty(len(self._invariants), dtype=np.promote_types(moments.vector.dtype, float))
        jac = None
        if get_jacobian:
            jac = np.zeros((len(self._invariants), len(moments)), dtype=vector.dtype)

        for idx, inv in enumerate(self._invariants):
            vector[idx] = inv.evaluate(moments)

            if get_jacobian:
                # Evaluate the derivatives
                for index, dphi in inv.get_gradient().items():
                    in_index = moments.linear_index(index)
                    jac[idx, in_index] = dphi.evaluate(moments.array)

        if self._real:
            vector = vector.real

        if get_jacobian:
            # Don't take 'real' of the Jacobian as if the moments are complex then the Jacobian
            # should be complex even though our output values are real
            return vector, jac

        return vector


def apply_invariants(invariants: List[MomentInvariant], moms: np.array, normalise=False, results=None) -> list:
    """Calculate the moment invariants for a given set of moments

    :param invariants: a list of invariants to calculate
    :param moms: the moments to use
    :param normalise: if True fill normalise the moments using the 0th moment
    :param results: an optional container to place the result in, if not supplied one will be created
    """
    if results is None:
        results = [None] * len(invariants)
    else:
        if not len(results) == len(invariants):
            raise ValueError('Results container must be of the same length as invariants')

    for idx, invariant in enumerate(invariants):
        results[idx] = invariant.apply(moms, normalise=normalise)

    return results


def read_invariants(filename: str = GEOMETRIC_INVARIANTS, read_max: int = None) -> \
        List[MomentInvariant]:
    """Read invariants in the format use by Flusser, Suk and Zitová.

    :param filename: the filename to read from, default to geometric moments invariants
    :param read_max: the maximum number of invariants to read
    """
    try:
        filename = INVS_MAP[filename]
    except KeyError:
        # Assume it is a filename
        pass

    invariants = []
    with open(filename, 'r') as file:

        for line in file:
            line = line.rstrip()
            if line:
                # New entry
                header = [int(number) for number in line.split(' ')]
                degree = header[2]
                builder = InvariantBuilder(degree)

                # Now read the actual terms
                line = file.readline().rstrip()
                while line:
                    terms = tuple(map(str_to_number, line.split(' ')))

                    prefactor = terms[0]

                    indices = []
                    # Extract the indices 3 at a time
                    for idx in range(degree):
                        indices.append(tuple(terms[idx * 3 + 1:(idx + 1) * 3 + 1]))
                    builder.add_term(prefactor, indices)

                    line = file.readline().rstrip()

                invariants.append(builder.build())
                if len(invariants) == read_max:
                    break

    return invariants


def str_to_number(value: str) -> Union[int, float, complex]:
    """Convert an integer, float or complex number string to the correct number type"""
    try:
        return int(value)
    except ValueError:
        pass

    try:
        return float(value)
    except ValueError:
        pass

    try:
        return complex(value)
    except ValueError:
        pass

    raise ValueError('{} is not an int, float or complex'.format(value))


def read(filename: str = GEOMETRIC_INVARIANTS, read_max: int = None, max_order=None) -> \
        MomentInvariants:
    """Read the invariants from file"""
    invariants = MomentInvariants()
    try:
        filename = INVS_MAP[filename]
    except KeyError:
        # Assume it is a filename
        pass

    if filename == COMPLEX_INVARIANTS:
        invariants.dtype = complex

    for inv in read_invariants(filename, read_max):
        if max_order is None or inv.max_order <= max_order:
            invariants.append(inv)
    return invariants


def calc_moment_invariants(
    invariants: Sequence[MomentInvariant],
    positions: np.array,
    sigma: Union[float, np.array] = 0.4,
    masses: Union[float, np.array] = 1.,
    normalise=False
) -> Sequence[float]:
    """Calculate the moment invariants for a set of Gaussians at the given positions."""
    max_order = 0

    # Calculate the maximum order invariant we'll need
    for inv in invariants:
        max_order = max(max_order, inv.max_order)

    raw_moments = geometric.from_gaussians(max_order, positions, sigma, masses)
    return tuple(invariant.apply(raw_moments, normalise) for invariant in invariants)



# -*- coding: utf-8 -*-
from . import analytic
from . import atomic
from . import dat
from . import descriptors
from . import exceptions
from . import functions
from . import generate
# from . import generative
from . import geometric
from . import invariants
from . import mathutil
from . import models
from . import optimisers
from . import polynomials
from . import utils
from . import play
from . import plot
from . import reconstruct
from . import sph
from . import zernike
from .play import asetools
from .atomic import *
from .fingerprinting import *
from .geometric import *
from .invariants import *
from .reconstruct import *
from .zernike import *

__all__ = (
              'analytic',
              'atomic',
              'invariants',
              'functions',
              'geometric',
              'mathutil'
              'utils',
              'play',
              'plot',
              'reconstruct',
              'exceptions',
              'models',
              'optimisers',
              'generate',
              'asetools',
              'sph',
              # 'generative',
              'polynomials',
              'descriptors',
              'dat',
              'invariants_'
              'zernike',
          ) \
          + zernike.__all__ + fingerprinting.__all__ + atomic.__all__ + invariants.__all__ + \
          geometric.__all__



# -*- coding: utf-8 -*-
import shutil

import ase
import ase.build
import numpy as np
import scipy.special
import pytest

from milad import asetools
from milad import atomic
from milad import descriptors
from milad import generate
from milad import functions
from milad import fingerprinting
from milad import invariants
from milad import utils
from milad import zernike

MILAD_TO_AMP = 52.63789013914202
"""Multiply by this strange factor that comes about because of minor differences between the way that we perform
the calculations"""
MILAD_TO_AMP_RADIAL = 1. / np.sqrt(3.)


def test_compare_zernike_with_amp():
    """Make sure we give the same results for the Zernike descriptor as AMP"""
    # pylint: disable=too-many-locals

    pytest.importorskip('amp')
    import amp.descriptor.zernike
    label = 'amp-fp'
    milad_label = 'amp-milad-fp'
    # Remove any files from the last test
    shutil.rmtree(f'./{label}-fingerprints.ampdb', ignore_errors=True)
    shutil.rmtree(f'./{label}-neighborlists.ampdb', ignore_errors=True)
    shutil.rmtree(f'./{milad_label}-fingerprints.ampdb', ignore_errors=True)
    shutil.rmtree(f'./{milad_label}-neighborlists.ampdb', ignore_errors=True)

    # Settings
    nmax = 7
    natoms = 6
    rcut = 2.8
    species = ('Si',)
    specie_numbers = (14,)
    specie_weight = 14.
    pts = generate.random_points_in_sphere(natoms, radius=1.5)
    atoms = atomic.AtomsCollection(natoms, positions=pts, numbers=specie_numbers[0])

    # Create the descriptors
    amp_descriptor = amp.descriptor.zernike.Zernike(
        cutoff=rcut, Gs=dict(Si=dict(Si=specie_weight)), nmax=nmax, elements=species, fortran=False, dblabel=label
    )

    # Create the MILAD descriptor
    milad_descriptor = descriptors.AmpDescriptor(
        fingerprinting.descriptor(
            features=dict(type=functions.WeightedDelta, kwargs=dict(weight=specie_weight)),
            moments_calculator=zernike.ZernikeMomentsCalculator(7, use_direct=True),
            invs=invariants.PowerSpectrum(mix_radials=False, radials_first=True),
            cutoff=rcut,
            apply_cutoff=True,
            smooth_cutoff=True
        ),
        dblabel='amp-milad-fp'
    )

    # Use AMP to calculate fingerprints
    system_dict = {'0': asetools.milad2ase(atoms)}
    amp_descriptor.calculate_fingerprints(system_dict, parallel=dict(cores=1), calculate_derivatives=False)

    milad_descriptor.calculate_fingerprints(system_dict, parallel=dict(cores=1), calculate_derivatives=False)

    for amp_fp, milad_fp in zip(amp_descriptor.fingerprints['0'], milad_descriptor.fingerprints['0']):
        assert amp_fp[0] == milad_fp[0]
        assert np.allclose(np.array(amp_fp[1]), MILAD_TO_AMP * np.array(milad_fp[1]))


def test_derivatives(complex_invariants):
    pytest.importorskip('amp')
    import amp.descriptor.zernike
    label = 'amp-fp'
    # Remove any files from the last test
    shutil.rmtree(f'./{label}-fingerprints.ampdb', ignore_errors=True)
    shutil.rmtree(f'./{label}-fingerprint-primes.ampdb', ignore_errors=True)
    shutil.rmtree(f'./{label}-neighborlists.ampdb', ignore_errors=True)

    # Settings
    nmax = 7
    rcut = 3.0
    species = ('Si',)
    specie_weight = 14.

    # Create the descriptors
    amp_descriptor = amp.descriptor.zernike.Zernike(
        cutoff=rcut, Gs=dict(Si=dict(Si=specie_weight)), nmax=nmax, elements=species, fortran=False, dblabel=label
    )

    # Create the MILAD descriptor
    milad_descriptor = descriptors.AmpDescriptor(
        fingerprinting.descriptor(
            features=dict(type=functions.WeightedDelta, kwargs=dict(weight=specie_weight)),
            moments_calculator=zernike.ZernikeMomentsCalculator(7),
            invs=complex_invariants,
            cutoff=rcut,
            apply_cutoff=True,
            smooth_cutoff=True
        ),
        dblabel='amp-milad-fp'
    )

    systems = {'0': ase.build.bulk('Si', 'hcp', a=2.6)}
    amp_descriptor.calculate_fingerprints(systems, parallel=dict(cores=1), calculate_derivatives=True)
    milad_descriptor.calculate_fingerprints(systems, parallel=dict(cores=1), calculate_derivatives=True)

    # Make sure we include the same interaction terms
    assert (set(amp_descriptor.fingerprintprimes['0']) - set(milad_descriptor.fingerprintprimes['0'])) == set()


def test_amp_zernike_comparison():
    pytest.importorskip('amp')
    import amp.descriptor.zernike
    label = 'amp-fp'
    # Remove any files from the last test
    shutil.rmtree(f'./{label}-fingerprints.ampdb', ignore_errors=True)
    shutil.rmtree(f'./{label}-neighborlists.ampdb', ignore_errors=True)

    nmax = 7
    natoms = 5
    rcut = 1.0

    pts = generate.random_points_in_sphere(natoms - 1)
    atoms = atomic.AtomsCollection(natoms - 1, positions=pts, numbers=14)

    milad_descriptor = functions.Chain(
        atomic.FeatureMapper(), functions.CosineCutoff(rcut), zernike.ZernikeMomentsCalculator(nmax, use_direct=True),
        invariants.PowerSpectrum(mix_radials=False, radials_first=True)
    )

    amp_descriptor = amp.descriptor.zernike.Zernike(
        cutoff=rcut, Gs=dict(Si=dict(Si=1.)), nmax=nmax, elements=['Si'], fortran=False, dblabel=label
    )

    system_dict = {'0': ase.Atoms(positions=np.concatenate((np.array([[0., 0., 0.]]), pts)), numbers=[14] * natoms)}
    amp_descriptor.calculate_fingerprints(system_dict, parallel=dict(cores=1), calculate_derivatives=False)

    phi = milad_descriptor(atoms)

    assert np.allclose(np.array(amp_descriptor.fingerprints['0'][0][1]), phi * MILAD_TO_AMP)


def test_radial_function():
    """Check that our Zernike radial function agrees with AMP"""
    # pylint: disable=invalid-name
    pytest.importorskip('amp')
    import amp.descriptor.zernike

    fac = scipy.special.factorial

    nmax = 7
    rho = 0.6353

    # Cache used by AMP
    factorial = [fac(0.5 * _) for _ in range(4 * nmax + 3)]

    for n in utils.inclusive(nmax):
        for l in utils.inclusive(n):
            amp_radial = amp.descriptor.zernike.calculate_R(n, l, rho, factorial)
            our_radial = zernike.r_nl(n, l, rho)
            assert np.isclose(our_radial, MILAD_TO_AMP_RADIAL * amp_radial)



# -*- coding: utf-8 -*-
from . import interfaces
from . import powerspectrum
from . import moment_invariants
from . import invertible_invariants
from .moment_invariants import *
from .powerspectrum import *
from .interfaces import *

__all__ = powerspectrum.__all__ + interfaces.__all__ + moment_invariants.__all__ + ('invertible_invariants',)



# -*- coding: utf-8 -*-
import functools
import collections
import logging
from typing import List

import numpy as np
import rmsd as rmsdlib
from scipy.spatial import distance
from scipy import optimize
from sklearn import cluster

from . import atomic
from . import base_moments
from . import exceptions
from . import fingerprinting
from . import mathutil
from . import optimisers

_LOGGER = logging.getLogger(__name__)

StructureOptimisationResult = collections.namedtuple('StructureOptimisationResult', 'success atoms message rmsd')


@functools.singledispatch
def find_clusters(spec, num_clusters: int, **kwargs) -> np.ndarray:
    """Find clusters based on moments"""
    raise TypeError(f'Cannot find clusters from {spec.__class__.__name__}')


@find_clusters.register(base_moments.Moments)
def _(moments: base_moments.Moments, num_clusters: int, query: base_moments.ReconstructionQuery,
      descriptor: fingerprinting.MomentInvariantsDescriptor) -> np.ndarray:
    """Find clusters from moments.  This will take the moments and reconstruct values on a grid
    which will be used for the actual cluster determination"""
    # Calculate the grid values
    values = moments.reconstruct(query, zero_outside_domain=True)
    # Find the cluster centres
    return find_clusters((query.points, values), num_clusters)


@find_clusters.register(tuple)
def _(grid, num_clusters: int) -> np.ndarray:
    """Given a grid consisting of points and weights this function funds clusters using k-means"""
    grid_points, grid_values = grid
    kmeans = cluster.KMeans(num_clusters, max_iter=1000, algorithm='full', tol=1e-1)
    kmeans.fit(grid_points, sample_weight=grid_values)

    num_labels = len(set(kmeans.labels_))
    if num_labels != num_clusters:
        raise exceptions.ReconstructionError(f'Could not find {num_clusters} clusters, found {num_labels} labels')
    if len(kmeans.cluster_centers_) != num_clusters:
        raise exceptions.ReconstructionError(
            f'Could not find {num_clusters} clusters, found {len(kmeans.cluster_centers_)} centres'
        )

    return kmeans.cluster_centers_


def get_surrounding_gridpoints(
    query: base_moments.ReconstructionQuery, atom_pos: np.array, radius: float
) -> np.ndarray:
    # First find the points within a cube of the central point
    mask = (query.points[:, 0] > atom_pos[0] - radius) & \
           (query.points[:, 0] < atom_pos[0] + radius) & \
           (query.points[:, 1] > atom_pos[1] - radius) & \
           (query.points[:, 1] < atom_pos[1] + radius) & \
           (query.points[:, 2] > atom_pos[2] - radius) & \
           (query.points[:, 2] < atom_pos[2] + radius)

    indices = np.argwhere(mask)[:, 0]
    rsq = radius * radius

    # Now check the gridpoints left to see if they fall within the cutoff sphere
    for idx in indices:
        dr = query.points[idx] - atom_pos  # pylint: disable=invalid-name
        if np.dot(dr, dr) > rsq:
            # Remove it from the mask
            mask[idx] = False

    return mask


def get_buffer_indices(query: base_moments.ReconstructionQuery, grid_values) -> List[int]:
    # Get all grid values that are currently 0, these should stay zero
    outside = list(np.argwhere(grid_values == 0.).reshape(-1))

    # Now, get those that are near the bounds of the sphere i.e. 1.0
    cutoff_sq = 0.85**2
    for idx, pt in enumerate(query.points):  # pylint: disable=invalid-name
        if np.sum(pt**2) > cutoff_sq:
            outside.append(idx)

    return outside


def find_atoms(
    # pylint: disable=too-many-locals
    num_atoms: int,
    moments: base_moments.Moments,
    descriptor: fingerprinting.MomentInvariantsDescriptor,
    query: base_moments.ReconstructionQuery = None,
    grid_size=31,
) -> atomic.AtomsCollection:
    """
    Given a set of moments this will try to extract atom positions and species by successively placing a new atom
    on the current peak in reconstruction grid.  After this the a new grid is calculated and subtracted from the
    original.

    :param moments:
    :param descriptor:
    :param num_atoms: the number of atoms to find
    :param query: the reconstruction query to use.
    :param grid_size: the size of reconstruction grid to use.  Only used if no reconstruction query is passed.
    :return:
    """
    query = query or moments.create_reconstruction_query(moments.get_grid(grid_size), moments.max_order)

    atom_positions = []
    atom_numbers = []
    optimiser = optimisers.StructureOptimiser()

    # Calculate the original moments grid
    orig_grid = moments.reconstruct(query, zero_outside_domain=True)
    current_grid = orig_grid

    preprocess = False

    while True:
        # Find the index of the maximum value in the current grid
        max_idx = current_grid.argmax()

        # Get that position in the grid
        atom_positions.append(query.points[max_idx] / descriptor.cutoff)
        atom_numbers.append(1.)

        # Build an atoms collection with the current set of atoms
        current_atoms = atomic.AtomsCollection(len(atom_positions), positions=atom_positions, numbers=atom_numbers)

        # Locally optimise the atomic positions
        res = optimiser.optimise(
            descriptor,
            target=moments,
            initial=current_atoms,
            preprocess=preprocess,
        )

        if num_atoms == current_atoms.num_atoms:
            # Reached the number of atoms limit
            break

        # Get the moments so we can subtract this from the grid
        current_moments = descriptor.get_moments(res.value, preprocess=preprocess)

        # Get the grid for this set of atoms
        this_grid = current_moments.reconstruct(query, zero_outside_domain=True)

        # Subtract off the current grid
        current_grid = orig_grid - this_grid

        # Now remove the signal of the atom from the grid
        remove_idxs = np.argwhere(current_grid >= (0.5 * current_grid.max()))
        current_grid[remove_idxs] = 0.

        atom_positions = list(res.value.positions)
        atom_numbers = list(res.value.numbers)

    return res.value


def create_atoms_collection(clusters: cluster.KMeans, atomic_numbers=1.):
    """Take a set of clusters and use the centres to construct an atoms collection"""
    num_atoms = len(clusters.cluster_centers_)
    return atomic.AtomsCollection(num_atoms, positions=clusters.cluster_centers_, numbers=atomic_numbers)


DecoderResult = collections.namedtuple(
    'DecoderResult', 'success message value rmsd moments_reconstruction initial_reconstruction atoms_reconstruction'
)


def merge_atoms(system: atomic.AtomsCollection, dist_threshold=0.2):
    # Proceed to merging of atoms
    dists = distance.cdist(system.positions, system.positions)
    np.fill_diagonal(dists, np.inf)  # Get rid of diagonals as this is just the self-interaction
    merge_sets = []
    for (i, j) in np.argwhere(dists < dist_threshold):
        if i > j:
            # Ignore lower diagonal
            continue

        merged = False
        for merge_set in merge_sets:
            if i in merge_set or j in merge_set:
                if i in merge_set:
                    merge_set.add(j)
                elif j in merge_set:
                    merge_set.add(i)
                merged = True
                break

        if not merged:
            # Create a new merge set
            merge_sets.append({i, j})

    # Start merging
    positions = []
    numbers = []
    merged_indices = set()
    for merge_set in merge_sets:
        pos = np.zeros(3)
        number = 0.
        for i in merge_set:
            pos += system.positions[i]
            number += system.numbers[i]
            merged_indices.add(i)
        pos /= len(merge_set)

        positions.append(pos)
        numbers.append(number)

    for i in set(range(len(system.numbers))) - merged_indices:
        positions.append(system.positions[i])
        numbers.append(system.numbers[i])

    return len(merged_indices), atomic.AtomsCollection(len(positions), positions, numbers)


def get_best_rms(
    reference: atomic.AtomsCollection,
    probe: atomic.AtomsCollection,
    max_attempts: int = 1000,
    max_retries=20,
    threshold=1e-7,
    use_hungarian=True
) -> float:
    """
    Get the best RMSs fitting between two molecules.  This will first use an algorithm to make a decent guess at the
    best permutational ordering of atoms and then try a brute force search.

    :param reference: the reference to fit to
    :param probe: the probe to try to fit to the reference
    :param max_attempts: the number of attempts that RDKit should make the find the best RMSD in each retry
    :param max_retries: the number of ties we should call RDKit's get best RMSD method
    :param threshold: if the RMSD returned by RDKit's get best RMSD method drops below this value then we stop and
        return the result
    :param use_hungarian: if True will use the Hungarian algorithm (https://en.wikipedia.org/wiki/Hungarian_algorithm)
        to try and find a good label assignment before calling RDKit
    :return: the best RMSD found
    """
    from rdkit.Chem import rdMolAlign
    from . import rdkittools

    if use_hungarian:
        # Find a decent re-ordering of the atoms
        try:
            reorder_map = rmsdlib.reorder_hungarian(
                reference.numbers, probe.numbers, reference.positions, probe.positions
            )
        except ValueError:
            pass
        else:
            probe = atomic.AtomsCollection(
                reference.num_atoms,
                positions=probe.positions[reorder_map],
                numbers=probe.numbers[reorder_map],
            )

    # Now ask rdkit to find the best.
    # This will try a brute-force permutation search starting with the one we just defined which is likely to be
    # the best
    reference = rdkittools.milad2rdkit(reference)
    probe = rdkittools.milad2rdkit(probe)

    try:
        best = np.inf
        for _ in range(max_retries):
            best = min(rdMolAlign.GetBestRMS(probe, reference, maxMatches=max_attempts), best)
            if best < threshold:
                break
        return best
    except RuntimeError:
        return np.inf


def find_iteratively(
    descriptor: fingerprinting.MomentInvariantsDescriptor,
    fingerprint: np.ndarray,
    initial: atomic.AtomsCollection,
    find_species=False,
    min_rmsd=1e-7,
    max_iters=6,
    grid_query=None,
    structure_optimiser=None,
    minsep=0.55,
    verbose=False,
) -> optimisers.StructureOptimisationResult:
    """
    Find a set of atoms from a fingerprint using an iterative approach.

    This algorithm proceeds in iterations as follows:
        1. Gradient descent to find the moments from the invariants (using `initial` atoms as the starting point)
        2. From the found moments:
            i) Identify the peaks in the grid sampled from the reconstructed moments
            ii) Create a collection of atoms and gradient descent to find the minimum of the atomic degrees of freedom
                wrt the moments
        3. Take the found atoms collection and perform gradient descent of the atomic degrees of freedom wrt to the
           original fingerprint.

    This is then repeated until either the RMSD between the fingerprint of the found atoms collection and the target
    drops below the passed threshold or the max number of iterations is reached.

    :param descriptor: the descriptor to use
    :param fingerprint: the target fingerprint
    :param initial: the initial atoms collection to start from
    :param find_species: if True, will look gradient descent atomic species as well as atomic coordinates
    :param min_rmsd: the fingerprint stopping threshold
    :param max_iters: the maximum number of iterations
    :param grid_query: an optional grid query, this is used when supersampling the found moments and if this function
        is being called many times using the same descriptor it is much faster to create the query externally and pass
        it in each time
    :param structure_optimiser: an optional structure optimiser to use
    :param minsep: the minimum separation between atoms
    :param verbose: if True, will print information about progress
    :return: the result from the last structure optimisation
    """
    # pylint: disable=too-many-locals, too-many-branches
    # Initialisation
    num_atoms = initial.num_atoms
    moments_optimiser = optimisers.MomentsOptimiser()
    if structure_optimiser is None:
        structure_optimiser = optimisers.StructureOptimiser()
        if minsep:
            structure_optimiser.separation_force = atomic.SeparationForce(epsilon=1e-1, cutoff=minsep, power=2)

    atoms = initial
    mask = None
    if not find_species:
        # Fix the species numbers
        mask = atoms.get_mask()
        mask.numbers = atoms.numbers

    for i in range(max_iters):
        # Now recreate the moments from the atoms
        moments = descriptor.get_moments(atoms)

        if grid_query is None:
            # Create the reconstruction query the first time
            grid_query = moments.create_reconstruction_query(moments.get_grid(31), moments.max_order)

        if verbose:
            print(f'{i}: Finding moments from fingerprint...', end='')
        result = moments_optimiser.optimise(
            invariants_fn=descriptor.invariants,
            target=fingerprint,
            initial=moments,
            verbose=(verbose == 'high'),
        )
        moments = result.value
        if verbose:
            print(f'rmsd {result.rmsd}')

        if verbose:
            print(f'{i}: Finding atoms from moments...', end='')
        result = find_atoms_from_moments(
            descriptor,
            moments,
            num_atoms,
            numbers=atoms.numbers,
            minsep=minsep,
            mask=mask,
            grid_query=grid_query,
            structure_optimiser=structure_optimiser,
            verbose=(verbose == 'high')
        )
        atoms = result.value
        if verbose:
            print(f'rmsd {result.rmsd}')

        if verbose:
            print(f'{i}: Finding atoms from fingerprint...', end='')
        result = structure_optimiser.optimise(
            descriptor,
            target=fingerprint,
            initial=atoms,
            mask=mask,
            verbose=(verbose == 'high'),
        )
        if verbose:
            print(f'rmsd {result.rmsd}')

        if find_species:
            # Take the current result, fix the species and allow positions to vary
            atoms = result.value

            pos_mask = atoms.get_mask()
            pos_mask.numbers = atoms.numbers

            if verbose:
                print(f'{i}: Optimising atomic positions wrt fingerprint...', end='')
            result = structure_optimiser.optimise(
                descriptor,
                target=fingerprint,
                initial=atoms,
                mask=pos_mask,
                verbose=(verbose == 'high'),
            )
            if verbose:
                print(f'rmsd {result.rmsd}')

        if result.rmsd < min_rmsd:
            break

        atoms = result.value

    return result


# Interface for atoms finder:
# (descriptor, moments, num_atoms: int, *, verbose: Bool)


def find_atoms_from_moments(
    descriptor: fingerprinting.MomentInvariantsDescriptor,
    moments,
    num_atoms: int,
    *,
    numbers=1.,
    mask=None,
    grid_query=None,
    structure_optimiser=None,
    minsep=0.51,
    verbose=False,
):
    # Find the peaks and create the corresponding collection of atoms
    if verbose:
        print('Finding peaks...', end='')
    peaks = find_peaks(descriptor, moments, num_atoms, query=grid_query, subtract_signal=True, exclude_radius=minsep)
    atoms = atomic.AtomsCollection(num_atoms, peaks, numbers=numbers)
    if verbose:
        print(f'found {atoms.num_atoms}')

    if structure_optimiser is None:
        structure_optimiser = optimisers.StructureOptimiser()
        structure_optimiser.separation_force = atomic.SeparationForce(epsilon=1e-2, cutoff=minsep, power=1)
    result = structure_optimiser.optimise(
        descriptor,
        target=moments,
        initial=atoms,
        mask=mask,
        verbose=verbose,
    )

    return result


def find_maximum(
    descriptor: fingerprinting.MomentInvariantsDescriptor,
    moments: base_moments.Moments,
    num_samples=128,
) -> optimize.OptimizeResult:
    """Given a set of moments this function will use a function minimisation algorithm to find
    the global maximum of the moments reconstruction

    :param descriptor: the descriptor that created the moments
    :param moments: the moments
    :param num_samples: the number of samples to use for the first pass of the global maximum search
    """
    # pylint: disable=invalid-name
    cutoff = descriptor.cutoff

    def adapter(moms, pt):
        return -moms.value_at(np.array([mathutil.sph2cart(pt)]))

    res = optimize.shgo(
        lambda pt: adapter(moments, pt),
        bounds=[(0.0, 1.), (0., np.pi), (0., 2 * np.pi)],
        n=num_samples,
        sampling_method='sobol'
    )
    # Convert coordinates back to cartesian and scale
    res.update(dict(x=mathutil.sph2cart(res.x) / cutoff, xl=mathutil.sph2cart(res.x) / cutoff))
    return res


def find_peaks(
    descriptor: fingerprinting.MomentInvariantsDescriptor,
    moments: base_moments.Moments,
    num: int,
    *,
    query: base_moments.ReconstructionQuery = None,
    grid_size=31,
    exclude_radius=0.51,
    subtract_signal=True
):
    query = query or moments.create_reconstruction_query(moments.get_grid(grid_size), moments.max_order)
    current_grid = moments.reconstruct(query, zero_outside_domain=True)
    return find_peaks_from_grid(
        descriptor, num, current_grid, query, exclude_radius=exclude_radius, subtract_signal=subtract_signal
    )


def find_peaks_from_grid(
    descriptor: fingerprinting.MomentInvariantsDescriptor,
    num_peaks: int,
    grid_values: np.ndarray,
    query: base_moments.ReconstructionQuery,
    exclude_radius=0.51,
    subtract_signal=True
):
    # pylint: disable=too-many-locals
    scale = 1.0 if descriptor.cutoff is None else 1. / descriptor.cutoff

    current_grid = grid_values.copy()

    exclude = exclude_radius * scale
    outside = get_buffer_indices(query, current_grid)

    mask = np.zeros(current_grid.shape, dtype=bool)
    mask[outside] = True

    found_positions = []
    for _ in range(num_peaks):
        # Find the index of the maximum value in the current grid
        max_idx = current_grid.argmax()

        # Get that position in the grid
        atom_pos = query.points[max_idx]
        found_positions.append(atom_pos / scale)

        if subtract_signal:
            # Build an atoms collection with a single atom at that position
            single_atom = atomic.AtomsCollection(1, positions=[atom_pos / scale], numbers=[1.])

            # Get the moments so we can subtract this from the grid
            single_moments = descriptor.get_moments(single_atom, preprocess=False)

            # Get the grid for just that atom on its own
            atom_grid = single_moments.reconstruct(query, zero_outside_domain=True)

            # Subtract off the single atom grid
            current_grid -= atom_grid

        # Now mask off the gridpoints associated with this atom
        mask |= get_surrounding_gridpoints(query, atom_pos, exclude)

        current_grid[mask] = 0.

    return np.array(found_positions)



# -*- coding: utf-8 -*-
import numpy as np

from milad import functions
from milad import base_moments

__all__ = ('Invariants',)


class Invariants(functions.Function):
    # pylint: disable=abstract-method
    input_type = base_moments.Moments
    output_type = np.ndarray



# -*- coding: utf-8 -*-
from milad import dat


def test_fingerprint_set(fingerprint_set: dat.FingerprintSet, descriptor, training_data):
    assert descriptor.fingerprint_len == fingerprint_set.fingerprint_len
    assert fingerprint_set.has_all_forces()
    assert fingerprint_set.has_all_derivatives()

    total_envs = sum([len(atoms) for atoms in training_data])
    assert total_envs == fingerprint_set.total_environments

    # Test fingerprints appear as we would expect
    assert len(training_data) == len(fingerprint_set.fingerprints)
    for atoms, fps in zip(training_data, fingerprint_set.fingerprints):
        # One FP per atom
        assert len(fps) == len(atoms)

    assert all([len(atoms) == size for atoms, size in zip(training_data, fingerprint_set.sizes)])

    # Potential energies
    for atoms, energy in zip(training_data, fingerprint_set.get_potential_energies()):
        assert atoms.get_potential_energy() / len(atoms) == energy



# -*- coding: utf-8 -*-
import numpy
from scipy.spatial.transform import Rotation

import milad
from milad import transform
from milad import geometric


def test_general(geometric_invariants):
    # Settings
    scale = 1.
    sigma = scale * 0.4
    num_rotations = 0
    num_points = 4

    pos1 = 4 * numpy.random.rand(num_points, 3)
    pos1 = transform.center(pos1)
    pos1 *= scale

    moments = geometric.from_gaussians(geometric_invariants.max_order, pos1, sigma)
    invariants = geometric_invariants(moments)

    for rot in range(1, num_rotations + 1):
        rot = Rotation.random()
        rotated = rot.apply(pos1)

        # Calculate the moment invariants
        rot_invariants = milad.invariants.calc_moment_invariants(geometric_invariants, rotated, sigma)

        assert numpy.testing.assert_array_almost_equal(invariants, rot_invariants)



# -*- coding: utf-8 -*-
# import numpy as np
#
# from milad import reconstruct, fingerprinting, atomic
#
#  Skipping this for now
# def test_find_from_moments(complex_invariants):
#     num_atoms = 10
#
#     descriptor = fingerprinting.descriptor(invs=complex_invariants, cutoff=1.0)
#
#     atoms = atomic.random_atom_collection_in_sphere(num_atoms)
#     mask = atoms.get_mask()
#     mask.numbers[:] = 1.
#     moments = descriptor.get_moments(atoms)
#
#     res = reconstruct.find_atoms_from_moments(
#         descriptor,
#         moments,
#         num_atoms,
#         mask=mask,
#         verbose=True,
#     )
#
#     assert np.allclose(atoms.positions, res.value.positions)



# -*- coding: utf-8 -*-
import abc
import collections
import numbers
from typing import Tuple, Dict, Set

import numpy as np
import numba


class Polynomial(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def evaluate(self, values: np.ndarray):
        """Evaluate the polynomial with given set of values"""


class HomogenousPolynomial(Polynomial):
    """A homogenous polynomial.

    These are not strictly homogenous as they are allowed to have a numerical additive
    constant e.g. 5 x_0,0,0 + 2.  This makes them easier to work with and a better fit for our use case.
    """

    def __init__(
        self,
        degree: int,
        prefactors: np.ndarray = None,
        terms: np.ndarray = None,
        constant=0.,
        conjugate_values=False,
        simplify=True,
    ):
        """
        Create a new homogenous polynomial of a given degree

        :param degree: the degree of the polynomial
        :param prefactors: the array of prefactors
        :param terms: the array indices that make up this polynomial where the array is interpreted as
            [sum index, product index, value array index]
        :param constant: the additive constant
        """
        prefactors = np.array(prefactors)
        terms = np.array(terms)

        if terms.size == 0 and len(terms.shape) == 1:
            # Make terms the right shape if it is empty
            terms = np.empty((0, 0, 0))

        if len(terms.shape) != 3:
            raise ValueError(f'terms must be rank three array, got {len(terms.shape)}')
        if len(prefactors) != len(terms):
            raise ValueError(
                f'prefactors and terms must have same length, got prefactors={len(prefactors)}, terms={len(terms)}'
            )

        if simplify:
            prefactors, terms = self._simplify(prefactors, terms)

        self._degree = degree
        self._prefactors = prefactors
        self._terms = terms
        self._constant = constant
        self._conjugate = conjugate_values
        self._derivatives_cache = {}
        self._variables = None
        if self._terms.size > 0:
            self._evaluate_method = self._numpy_evaluate
        else:
            self._evaluate_method = lambda *args: self._constant

    def __mul__(self, value: numbers.Number):
        if isinstance(value, numbers.Number):
            if value == 1.:
                return self

            return HomogenousPolynomial(
                self._degree, self._prefactors * value if self._prefactors is not None else None,
                self._terms if self._terms is not None else None, self._constant * value
            )

        raise TypeError(f'Unexpected type {value.__class__.__name__}')

    __rmul__ = __mul__

    def __add__(self, other):
        if other == 0.:
            return self

        if not isinstance(other, HomogenousPolynomial):
            raise TypeError(f'Unexpected type {other.__class__.__name__}')

        if other._degree != self._degree:
            return ValueError(
                f'Cannot add two homogenous polynomials of different degree ({self.degree} vs {other.degree})'
            )

        return HomogenousPolynomial(
            self.degree,
            prefactors=np.concatenate((self._prefactors, other._prefactors)),
            terms=np.concatenate((self._terms, other._terms)),
            constant=self._constant + other.constant
        )

    __radd__ = __add__

    def __str__(self) -> str:
        if self._prefactors is None:
            return 'None'

        terms = [
            f"{prefactor} * x_{','.join(map(str, term))}" for prefactor, term in zip(self._prefactors, self._terms)
        ]
        if self._constant:
            terms.append(str(self._constant))

        return ' + '.join(terms)

    @property
    def degree(self) -> int:
        return self._degree

    @property
    def prefactors(self) -> np.ndarray:
        return self._prefactors

    @property
    def terms(self) -> np.ndarray:
        return self._terms

    @property
    def constant(self):
        return self._constant

    @property
    def variables(self) -> Set[Tuple]:
        """Return a set of all the the indices used by these invariants"""
        if self._variables is None:
            indices = set()
            for product in self._terms:
                for idx in product:
                    indices.add(tuple(idx))
            self._variables = indices

        return self._variables

    def conjugate(self):
        """Return the complex conjugate of this polynomial"""
        return HomogenousPolynomial(
            self._degree,
            self._prefactors.conjugate() if self._prefactors is not None else None,
            self._terms if self._terms is not None else None,
            self._constant.conjugate(),
            conjugate_values=not self._conjugate
        )

    def evaluate(self, values: np.ndarray):
        if self._conjugate:
            values = values.conjugate()

        if isinstance(values, np.ndarray):
            return self._numpy_evaluate(values)

        return self._generic_evaluate(values)

    def _numpy_evaluate(self, values: np.ndarray):
        """Fast method to get the invariant from a numpy array"""
        total = self._constant  # type: float

        if self._terms.size > 0:
            if values.dtype == object or len(values.shape) > 3:
                total += numpy_evaluate(self.prefactors, self._terms, values)
            else:
                total += numba_evaluate(self.prefactors, self._terms, values)

        return total

    def _generic_evaluate(self, values):
        """Generic apply for values that support indexing.

        This is slower version of above but compatible with values that aren't numpy arrays"""
        total = self._constant
        for factor, indices in zip(self._prefactors, self._terms):
            product = 1
            for index in indices:
                product *= values[tuple(index)]
            total += factor * product
        return total

    __call__ = evaluate

    def get_partial_derivative(self, variable) -> 'HomogenousPolynomial':
        variable = tuple(variable)
        try:
            return self._derivatives_cache[variable]
        except KeyError:
            pass

        prefactors = []
        terms = []
        constant = 0

        for prefactor, product in zip(self._prefactors, self._terms):
            # Gather the terms in this product
            powers = collections.defaultdict(int)
            for indices in product:
                powers[tuple(indices)] += 1

            try:
                power = powers.pop(variable)
            except KeyError:
                # This term will not be in the derivative
                pass
            else:
                # Calculate the new prefactor
                new_prefactor = prefactor * power

                new_product = []
                # And add in the correct multiple of this variable
                if power != 1:
                    new_product.extend((variable,) * (power - 1))

                for var, power in powers.items():
                    # The other terms keep their exponent unchanged
                    new_product.extend((var,) * power)

                if power == 1 and len(product) == 1:
                    constant += new_prefactor
                else:
                    prefactors.append(new_prefactor)
                    terms.append(new_product)

        new_degree = self.degree - 1 if prefactors else 0
        deriv = HomogenousPolynomial(new_degree, prefactors, terms, constant, conjugate_values=self._conjugate)

        self._derivatives_cache[variable] = deriv
        return deriv

    def get_gradient(self) -> Dict[Tuple, 'HomogenousPolynomial']:
        return {variable: self.get_partial_derivative(variable) for variable in self.variables}

    @staticmethod
    def _simplify(prefactors: np.ndarray, terms: np.ndarray):
        new_prefactors = []
        new_terms = []
        for prefactor, product in zip(prefactors.tolist(), terms.tolist()):
            if prefactor == 0:
                continue

            found = False
            for i, new_term in enumerate(new_terms):
                if product == new_term:
                    new_prefactors[i] += prefactor
                    found = True
                    break

            if not found:
                new_prefactors.append(prefactor)
                new_terms.append(product)

        if new_prefactors:
            new_terms, new_prefactors = tuple(zip(*sorted(zip(new_terms, new_prefactors))))

        return np.array(new_prefactors), np.array(new_terms)


def numpy_evaluate(prefactors, indices: np.array, values: np.ndarray):
    """Fast method to get the polynomial from a numpy array"""
    return np.dot(prefactors, np.prod(values[indices[:, :, 0], indices[:, :, 1], indices[:, :, 2]], axis=1))


@numba.jit(parallel=False, nopython=True)
def numba_evaluate(prefactors, indices, moments):
    """Numba version to speed up calculation of larger polynomials"""
    total = 0.
    for idx in numba.prange(len(prefactors)):  # pylint: disable=not-an-iterable
        factor = prefactors[idx]
        entry = indices[idx]

        product = factor
        for index in entry:
            product *= moments[index[0], index[1], index[2]]

        total += product
    return total


class PolyBuilder:
    """A convenience builder that will start a polynomial using __getitem__ so clients can just use:
        m = PolyBuilder()
        5 * m[3, 1, 3]
    to start a polynomial
    """

    def __getitem__(self, indices):
        return HomogenousPolynomial(
            1,
            prefactors=np.array([1.]),
            terms=np.array([[list(indices)]]),
        )



# -*- coding: utf-8 -*-
from . import interfaces
from . import amp_adapter
from . import _dscribe
from .interfaces import *
from ._dscribe import *
from .amp_adapter import *

__all__ = _dscribe.__all__ + amp_adapter.__all__ + interfaces.__all__



# -*- coding: utf-8 -*-
import numpy as np
import sympy

from milad import functions


def test_function(function: functions.Function, value, expected_output=None, check_jacobian=None):
    if function.input_type:
        assert issubclass(type(value), function.input_type)
    output = function(value)
    if function.output_type:
        assert issubclass(type(output), function.output_type)

    if expected_output is not None:
        assert np.all(functions.get_bare(output) == functions.get_bare(expected_output))

    if check_jacobian is None:
        check_jacobian = function.supports_jacobian

    # Try checking the inverse
    if function.inverse is not None:
        inverted = function.inverse(output)
        assert np.allclose(functions.get_bare(value), functions.get_bare(inverted))

    if check_jacobian:
        # Check the derivatives
        vector_input = functions.get_bare(value)
        reals = np.isreal(vector_input)
        # Create the symbolic inputs to the function
        input_vec = np.array([sympy.Symbol(f'x{i}', real=real) for i, real in enumerate(reals)])

        # Overwrite the input with a symbolic version
        if isinstance(value, np.ndarray):
            value = input_vec
        else:
            value.vector = input_vec

        output, jac = function(value, jacobian=True)
        output = functions.get_bare(output)
        if not isinstance(output, np.ndarray):
            # Must be a scalar
            output = [output]
        for i, out in enumerate(output):
            for j, variable in enumerate(input_vec):
                if np.ma.is_masked(jac[i, j]):
                    continue
                deriv = sympy.diff(out, variable)
                assert sympy_equal(jac[i, j], deriv)


def sympy_equal(expr1, expr2):
    if isinstance(expr1, sympy.Expr):
        expr1 = expr1.expand()
    expr2 = expr2.expand()
    difference = expr1 - expr2
    if not expr1 == expr2:
        if isinstance(difference, sympy.Number):
            return np.isclose(complex(difference), 0.)

        # If they differ, check that it's by a meaningful amount.
        # We cast to complex here, even for reals, because this will catch both cases without having to do checks
        coeffs = np.array(tuple(difference.as_coefficients_dict().values()), dtype=np.complex)
        return np.allclose(coeffs, 0.)

    return True


def generate_vectors_on_sphere(num):
    """Generate `num` random unit vectors on a sphere"""
    vecs = np.empty((num, 3))
    for i in range(num):
        vec = np.random.rand(3)
        vec /= np.linalg.norm(vec, axis=0)
        vecs[i] = vec
    return vecs



# -*- coding: utf-8 -*-
import numpy as np
from scipy.spatial.transform import Rotation

from milad import generate
from milad.invariants import powerspectrum
from milad import zernike

# pylint: disable=invalid-name


def invariants_test(invariants_fn):
    pts = generate.random_points_in_sphere(11)

    phi = invariants_fn(pts)

    for _ in range(10):
        random = Rotation.random()
        rotated = random.apply(pts)

        phi_prime = invariants_fn(rotated)

        assert np.allclose(phi, phi_prime)


def test_powerspectrum_basic():
    ps = powerspectrum.PowerSpectrum()

    def invariants_fn(pts):
        moments = zernike.from_deltas(n_max=7, positions=pts)
        return ps(moments)

    invariants_test(invariants_fn)



# -*- coding: utf-8 -*-
"""Module for things related to spherical harmonics"""
import numbers
from typing import Union, Tuple, Optional, NamedTuple

import numpy as np
import numpy.ma

from . import utils

# pylint: disable=invalid-name


class InclusiveRange(NamedTuple):
    min: numbers.Number
    max: numbers.Number

    def __contains__(self, x: numbers.Number):
        return self.min <= x <= self.max

    def __len__(self):
        return (self.max - self.min) + 1


Range = InclusiveRange
RangeType = Union[Tuple[int, int], Range]
MaxOrRange = Union[int, Tuple[type(None), type(None)]]

__all__ = 'IndexTraits', 'MaxOrRange', 'Range'


class IndexTraits:
    """Helper class to deal with various possible valid indices for spherical harmonics and a radial function"""

    def __init__(
        self, n_spec: MaxOrRange, l_spec: MaxOrRange = None, l_le_n: bool = False, n_minus_l_even: bool = False
    ):
        self._n_range = to_range(n_spec)
        self._l_range = to_range(l_spec or self._n_range[1])
        self._l_le_n = l_le_n
        self._n_minus_l_even = n_minus_l_even

        if l_le_n and self._n_range[1] > self._n_range[1]:
            raise ValueError(f'Cannot have l_max > n_max ({self._n_range[1]} > {self._n_range[1]})')

    @property
    def n(self) -> InclusiveRange:
        return self._n_range

    @property
    def N(self) -> int:
        """Return the total number of radial basis included"""
        return len(self._n_range)

    @property
    def l(self) -> InclusiveRange:
        return self._l_range

    @property
    def L(self) -> int:
        """Get the total number of angular frequencies included i.e. l = 0 -> L - 1"""
        return len(self._l_range)

    def iter_n(self, l: int, n_spec: MaxOrRange = None):
        """Iterate over all the valid n values for a given angular frequency, l"""
        n_min, n_max = make_range(n_spec, self._n_range)

        # If necessary, shift n_min up to the nearest higher or equal integer such that `even(n - l) is True`
        n_min = n_min + ((n_min - l) % 2) if self._n_minus_l_even else n_min
        yield from utils.inclusive(max(l, n_min) if self._l_le_n else 0, n_max, 2 if self._n_minus_l_even else 1)

    def iter_l(self, n: int, l_spec: MaxOrRange = None):
        """Iterate over all valid values of l for a given value of n"""
        l_min, l_max = make_range(l_spec, self._l_range)

        # If necessary, shift l_min up to the nearest higher or equal integer such that `even(n - l) is True`
        l_min = l_min + ((n - l_min) % 2) if self._n_minus_l_even else l_min
        yield from utils.inclusive(l_min, min(l_max, n) if self._l_le_n else l_max, 2 if self._n_minus_l_even else 1)

    @staticmethod
    def iter_m(l: int, m_spec: MaxOrRange = None):
        m_min, m_max = make_range(m_spec, (-l, l))
        yield from utils.inclusive(m_min, m_max)

    def iter_nl(self, n_spec: MaxOrRange = None, l_spec: MaxOrRange = None):
        n_spec = make_range(n_spec, self._n_range)
        l_spec = make_range(l_spec, self._l_range)

        for n in utils.inclusive(*n_spec, 1):
            for l in self.iter_l(n, l_spec):
                yield n, l

    def iter_nlm(self, n_spec: MaxOrRange = None, l_spec: MaxOrRange = None, m_spec: MaxOrRange = None):
        for n, l in self.iter_nl(n_spec, l_spec):
            for m in self.iter_m(l, m_spec):
                yield n, l, m

    def iter_lm(self, n: int, l_spec: MaxOrRange = None, m_spec: MaxOrRange = None):
        for l in self.iter_l(n, l_spec):
            for m in self.iter_m(l, m_spec):
                yield l, m

    __iter__ = iter_nlm


def make_range(spec: Optional[MaxOrRange], default: RangeType) -> Range:
    if spec is None:
        return default

    val_range = list(to_range(spec))
    # Use our n_range as defaults if either min/max are empty
    val_range[0] = default[0] if val_range[0] is None else val_range[0]
    val_range[1] = default[1] if val_range[1] is None else val_range[1]

    return Range(*val_range)


def to_range(spec: MaxOrRange) -> Range:
    if isinstance(spec, tuple):
        return Range(*spec)
    if isinstance(spec, (int, np.integer)):
        return Range(0, spec)

    raise ValueError(f"'{spec}' is not a valid value specification, takes integer (maximum) or 2-tuple (range)")


def create_array(indices: IndexTraits, dtype=complex):
    """Create an empty array that is sized according to the passed indices"""
    array = np.empty((indices.N, indices.L, 2 * indices.L + 1), dtype=dtype)
    # Create the mask and unmask the valid elements
    mask = np.ones(array.shape, dtype=int)
    for idx in indices.iter_nlm():
        mask[idx] = 0
    return numpy.ma.array(array, mask=mask)



# -*- coding: utf-8 -*-
import numpy as np
import pytest
from scipy.spatial.transform import Rotation

import milad
from milad.invariants import invertible_invariants
from milad import zernike
from milad import generate

N_MAX = 5
indices = milad.sph.IndexTraits(n_spec=N_MAX, n_minus_l_even=True, l_le_n=True)

# pylint: disable=redefined-outer-name


@pytest.fixture(scope='session')
def inv_invariants():
    """A fixture to create a set of invertible invariants"""
    return invertible_invariants.InvariantsGenerator.generate_all(indices)


def check_inverted_invariants(invariants: milad.invariants.MomentInvariants, target: np.array, inverted: np.ndarray):
    diff = target - inverted
    not_zero = np.argwhere(~np.isclose(diff, 0.)).ravel()
    assert not np.any(not_zero), f'First non-zero: {invariants[not_zero[0]]}'


def test_invertible_invariants_basics(inv_invariants):
    """Here we:
        1. Create a fingerprint from a random set of moments
        2. Invert the fingerprint to recover some corresponding moments
        3. Calculate the fingerprint from the inverted moments
        4. Assert that the two fingerprints match
    """
    # Create some random moments and calculate the fingerprint
    pts = generate.random_points_in_sphere(11, radius=1., centre=False)
    moments = zernike.from_deltas(indices.n.max, pts)
    phi = inv_invariants(moments)

    inverted = zernike.ZernikeMoments(indices.n.max, indices.l.max)
    # Perform inversion
    inv_invariants.invert(phi, inverted)
    assert not np.any(np.isnan(inverted.array))

    inverted_phi = inv_invariants(inverted)
    check_inverted_invariants(inv_invariants, phi, inverted_phi)


def test_invertible_invariants_symmetric():
    """Test invertible invariants for symmetric environments"""
    indices = milad.sph.IndexTraits(n_spec=12, l_spec=4, n_minus_l_even=True, l_le_n=True)
    inv_invariants = invertible_invariants.InvariantsGenerator.generate_all(indices)

    pts = np.array([[-0.5, 0, 0], [0.5, 0, 0]])
    # pts = np.array([[-0.5, -0.5, 0], [-0.5, 0.5, 0], [0.5, -0.5, 0], [0.5, 0.5, 0]])
    moments = zernike.from_deltas(indices.n.max, pts, l_max=indices.l.max)

    phi = inv_invariants(moments)

    inverted = zernike.ZernikeMoments(indices.n.max, indices.l.max)

    # with pytest.raises(ValueError):
    inv_invariants.invert(phi, inverted)

    assert not np.any(np.isnan(inverted.array))
    inverted_phi = inv_invariants(inverted)
    # For now we can't deal with this case...stay tuned
    check_inverted_invariants(inv_invariants, phi, inverted_phi)


def test_invertible_invariants_are_rotation_invariant(inv_invariants):
    """Check that the generated invariants are, indeed, invariant to rotation"""
    num_points = 10
    num_rotations = 10

    pts = generate.random_points_in_sphere(10, radius=1.)
    weights = np.random.rand(num_points)
    moments0 = zernike.from_deltas(N_MAX, pts, weights)
    phi0 = inv_invariants(moments0)
    for _ in range(num_rotations):
        rot = Rotation.random()
        rotated = rot.apply(pts)

        moments = zernike.from_deltas(N_MAX, rotated, weights)
        phi = inv_invariants(moments)

        assert np.allclose(phi0, phi)


def test_against_chiral_tetrahedra(inv_invariants, chiral_tetrahedra):
    minus, plus = chiral_tetrahedra
    minus_phi = inv_invariants(zernike.from_deltas(inv_invariants.max_order, minus)).real
    plus_phi = inv_invariants(zernike.from_deltas(inv_invariants.max_order, plus)).real

    assert np.allclose(minus_phi, plus_phi)



# -*- coding: utf-8 -*-
"""
This module contains methods and tools to work with and invert the invariants that are similar
to those proposed in the following:

Bandeira, A. S., Blum-Smith, B., Kileel, J., Perry, A., Weed, J., & Wein, A. S. (2017).
Estimation under group actions: recovering orbits from invariants. ArXiv, (June).
http://arxiv.org/abs/1712.10163

We use an adapted version of their algorithm to recover moments from invariants.
"""
import collections
import functools
import logging
from typing import Tuple

import numpy as np
import sympy
from sympy.physics.quantum import cg
from sympy.physics import wigner

from milad import mathutil
from milad import polynomials
from milad import sph
from milad import utils
from . import moment_invariants

_LOGGER = logging.getLogger(__name__)

# We use a lot of variable names that are convenient for mathematics but don't conform to the Google
# code style so just disable for this file
# pylint: disable=invalid-name
SQRT_TWO = 2**0.5


def q_matrix(l: int, direct_indexing=True) -> np.ndarray:
    """Create a Q matrix that transforms real spherical harmonics to complex ones"""
    size = 2 * l + 1
    Q_ = np.zeros((size, size), dtype=complex)
    Q_[0, 0] = SQRT_TWO
    for m in utils.inclusive(1, l):
        Q_[m, -m] = 1  # Top left
        Q_[-m, -m] = 1j  # Bottom left

        Q_[-m, m] = -1j * (-1)**m  # Top right
        Q_[m, m] = (-1)**m  # Bottom right

    if not direct_indexing:
        # Shift the array such that m = m' = 0 is in the middle of the matrix, i.e. at l, l
        Q_ = np.roll(Q_, (l, l), axis=(0, 1))

    return Q_ / SQRT_TWO


class InvertibleInvariants(moment_invariants.MomentInvariants):
    """A set of invertible invariants"""

    def __init__(
        self, degree_1: moment_invariants.MomentInvariants, degree_2: np.ma.masked_array,
        degree_3: moment_invariants.MomentInvariants, index_traits: sph.IndexTraits
    ):
        super().__init__(*degree_1, *degree_2.compressed(), *degree_3, are_real=False)
        self._degree_1 = degree_1
        self._degree_2 = degree_2
        self._degree_3 = degree_3
        self._index_traits = index_traits

    def invert(self, phi: np.array, moments_out):
        lmax = self._index_traits.l[1]

        used_invariants = set()
        moments_out.array.fill(float('nan'))

        used_invariants.update(self._invert_degree_1(phi, moments_out))

        # Let's go up in l until we find an I2 we can solve for
        l = 1
        while l <= lmax:
            gram, phi_indices = self.gram_matrix(phi, l=l)
            vectors, rank = self.get_vectors_from_gram(gram, l)
            self._place_vectors(moments_out, vectors, l)
            used_invariants.update(phi_indices)

            if rank > 0:
                # Now, let's get a 3rd degree invariant that have l1=l2=l3 so we can resolve the +/-I_3 degeneracy
                try:
                    sign, inv_idx = self._determine_sign(phi, moments_out, l)
                except RuntimeError:
                    _LOGGER.warning('Unable to determine sign for I_2, l=%i', l)
                else:
                    moments_out.array[:, l, :] = sign * moments_out.array[:, l, :]
                    used_invariants.add(inv_idx)

                break

            l += 1

        l += 1

        # Now we can continue using I3 only
        used_invariants.update(self._invert_degree_3(phi, l, moments_out))

        _LOGGER.info('Used %i out of %i invariants during inversion', len(used_invariants), len(phi))
        return moments_out

    def gram_indices(self, l: int) -> np.ndarray:
        """Get the indices of the invariants that correspond to the upper-right portion of the Gram matrix"""
        num_deg_1 = len(self._degree_1)
        indices = np.ma.masked_array(np.empty(self._degree_2.shape, dtype=int), mask=self._degree_2.mask)
        valid_invs = np.argwhere(self._degree_2 != None)  # pylint: disable=singleton-comparison
        indices[valid_invs[:, 0], valid_invs[:, 1], valid_invs[:, 2]] = (np.arange(0, len(valid_invs)) + num_deg_1)

        linear_indices = indices[l, :, :].compressed()
        return linear_indices

    def gram_matrix(self, phi: np.ndarray, l: int) -> Tuple[np.ndarray, np.ndarray]:
        """Get the Gram matrix for the corresponding angular frequency, l"""
        # Let's get the indices for the degree-2 invariants
        phi_indices = self.gram_indices(l)

        # Create the Gram matrix
        gram_size = sum(1 for _ in self._index_traits.iter_n(l))
        gram = np.zeros((gram_size, gram_size), dtype=complex)

        # Copy over the invariant values and symmetrise
        gram[np.triu_indices(gram_size)] = phi[phi_indices]
        ilower = np.tril_indices(gram_size, -1)
        gram[ilower] = gram.T[ilower]  # pylint: disable=unsubscriptable-object

        return gram, phi_indices

    def reduced_gram(self, phi: np.ndarray, moments, l: int):
        # Get the original gram matrix and then subtract off the subspaces that we have already found
        gram, gram_indices = self.gram_matrix(phi, l)
        gram_size = gram.shape[0]

        # Copy over the invariant values and symmetrise
        idx = 0
        for i in range(gram_size):
            for j in range(i, gram_size):
                # Subtract off what we've found so far
                gram[i, j] -= self[gram_indices[idx]](moments)
                idx += 1

        # Resymmetrise the Gram
        ilower = np.tril_indices(gram_size, -1)
        gram[ilower] = gram.T[ilower]  # pylint: disable=unsubscriptable-object

        return gram

    # @staticmethod
    # def get_vectors_from_gram(gram: np.array, l: int) -> Tuple[int, np.array]:
    #     """Given a Gram matrix that that is interpreted as an the SxS (where there are S radial functions) inner
    #     product space of the vectors at a particular angular frequency, l, this will solve for a set of possible
    #     vectors up to isomorphism.
    #     The rank of the Gram matrix and an array of the vectors are returned.
    #     """
    #     X1 = np.zeros((gram.shape[0], 2 * l + 1), dtype=complex)
    #     rank = np.linalg.matrix_rank(gram)
    #
    #     # Check if we have anything to decompose, if not just save ourselves the trouble and return now
    #     if not gram.nonzero() or rank == 0:
    #         return rank, X1
    #
    #     # Let's do a Cholesky and extract a set of compatible X vectors
    #     L = cholesky(gram)
    #     L = L[:, :rank]  # Collect entries up to rank
    #
    #     bound = int(math.floor(rank / 2))
    #     m_values = tuple(m for m in utils.inclusive(-bound, bound, 1) if not (mathutil.even(rank) and m == 0))
    #
    #     # Get a rotation matrix that makes the vectors respect the conjugate symmetry
    #     Q = q_matrix(l)[m_values, :][:, m_values]
    #     X1[:, m_values] = L @ Q
    #
    #     return rank, X1

    @staticmethod
    def get_vectors_from_gram(gram: np.array, l: int) -> Tuple[np.array, int]:
        """Given a Gram matrix that that is interpreted as an the SxS (where there are S radial functions) inner product
         pace of the vectors at a particular angular frequency, l, this will solve for a set of possible vectors up to
         isomorphism.
        The rank of the Gram matrix and an array of the vectors are returned.
        """
        vectors = np.zeros((gram.shape[0], 2 * l + 1), dtype=complex)

        u, s, _vh = np.linalg.svd(gram, hermitian=True)
        rank = int(np.sum(~np.isclose(s, 0.)))

        # Check if we have anything to decompose, if not just save ourselves the trouble and return now
        if rank == 0:
            return vectors, rank

        real_vectors = np.zeros((gram.shape[0], (2 * l + 1)), dtype=complex)
        # real_vectors = u[:, :rank] * s[:rank]**0.5
        real_vectors[:, :rank] = u[:, :rank] * s[:rank]**0.5

        # bound = int(math.floor(rank / 2))
        # m_values = tuple(m for m in utils.inclusive(-bound, bound, 1) if not (mathutil.even(rank) and m == 0))
        # m_values = tuple(utils.inclusive(-l, l, 1))

        # Get a rotation matrix that makes the vectors respect the conjugate symmetry
        # Q = q_matrix(l)[m_values, :][:, m_values]
        Q = q_matrix(l, direct_indexing=True)
        # Perform the rotation and multiply by the Clebsch-Gordan coefficient
        # vectors[:, m_values] = (2 * l + 1)**0.25 * real_vectors @ Q
        vectors = (2 * l + 1)**0.25 * real_vectors @ Q

        if rank == 1:
            # alpha, beta, gamma = sympy.symbols("alpha, beta, gamma", real=True)
            sp_rot = wigner.wigner_d(sympy.Integer(l), 0, sympy.pi / 4, sympy.pi / 4)
            rot = np.roll(np.array(sp_rot, dtype=complex), (-l, -l), axis=(0, 1))
            vectors = vectors @ rot

        return vectors, rank

    def _invert_degree_1(self, phi: np.array, moments_out) -> set:
        """Get the degree 1 invariants.  There is no inversion to be done here, we just copy over all the values
        corresponding to c_n0^0 which are all rotation invariants"""
        used_invariants = set()

        # Degree 1, these are easy, all invariants are moments themselves
        indices = tuple(self._index_traits.iter_n(l=0))
        num = len(indices)
        moments_out.array[indices, 0, 0] = phi[:num]
        used_invariants.update(range(num))

        return used_invariants

    def _place_vectors(self, moments, vectors, l: int):
        # m_range = tuple(utils.inclusive(-l, l))
        for i, n in enumerate(self._index_traits.iter_n(l)):
            # moments.array[n, l, m_range] = vectors[i, m_range]

            for m in utils.inclusive(0, l):
                moments[n, l, m] = vectors[i, m]
        return moments

    def _determine_sign(self, phi: np.ndarray, moments_out, l: int):
        # Find I3 that have l1=l2=l3
        res = np.array(self.find(lambda inv: inv.weight == 3 and np.all(inv.terms_array[:, :, 1] == l)))
        # Get the index of a non-zero invariant that we can use to determine if we should use +Q or -Q
        try:
            # Find one that is non-zero
            i3_nonzero = np.argwhere(~np.isclose(phi[res], 0))[0][0]
        except IndexError:
            raise RuntimeError(f'Unable to determine sign for I_2, l={l}')
        else:
            i3_nonzero = res[i3_nonzero]
            inv_res = self[i3_nonzero](moments_out)
            sign = inv_res / phi[i3_nonzero]
            if np.isclose(sign, 1):
                return 1, i3_nonzero
            if np.isclose(sign, -1.):
                return -1, i3_nonzero

            raise RuntimeError(f'The passed moments do not match I3 invariant (idx={i3_nonzero})')

    def _invert_degree_3(self, phi: np.array, l_start: int, moments_out) -> set:
        # pylint: disable=too-many-locals
        used_invariants = set()

        # Degree 3 - Now let's frequency march to recover the rest
        for l3 in utils.inclusive(l_start, self._index_traits.l[1], 1):
            vectors, used = self._invert_deg_3_l(phi, moments_out, l3)
            self._place_vectors(moments_out, vectors, l3)
            used_invariants.update(used)

        return used_invariants

    def _invert_deg_3_l(self, phi: np.array, moments, l3: int):
        # pylint: disable=too-many-locals
        used_invariants = set()

        n_values = tuple(self._index_traits.iter_n(l3))
        m_values = tuple(self._index_traits.iter_m(l3))
        vectors = np.empty((len(n_values), 2 * l3 + 1), dtype=moments.dtype)

        for n3_, n3 in enumerate(self._index_traits.iter_n(l3)):
            # Get the correct invariants
            i3_idx = np.array(self.find(functools.partial(degree_3_with_unknown, n3, l3)))

            if len(i3_idx) < 2 * l3 + 1:
                _LOGGER.warning("Don't have enough invariants to solve for n=%i, l=%i", n3, l3)

            # Create the matrices of the system to be solved
            coeffs = np.zeros((len(i3_idx), 2 * l3 + 1), dtype=complex)

            for i, inv_idx in enumerate(i3_idx):
                invariant = self[inv_idx]
                # Get just the first two terms (the third is the unknown)
                for j, m in enumerate(utils.inclusive(-l3, l3)):
                    mask = invariant.terms_array[:, 2, 2] == m
                    if np.any(mask):
                        prefactors = invariant.prefactors[mask]
                        indices = invariant.terms_array[:, 0:2, :][mask]
                        # pylint: disable=protected-access
                        coeffs[i, j] = polynomials.numpy_evaluate(prefactors, indices, moments)

            # The known invariants
            phis = phi[i3_idx]
            res = np.linalg.lstsq(coeffs, phis, rcond=None)
            if res[2] < (2 * l3 + 1):
                _LOGGER.warning('Rank deficiency (%i < %i) in n=%i, l=%i', res[2], (2 * l3 + 1), n3, l3)

            used_invariants.update(set(i3_idx))

            vectors[n3_, m_values] = res[0]

        return vectors, used_invariants


def degree_3_with_unknown(n: int, l: int, inv: moment_invariants.MomentInvariant):
    return inv.weight == 3 and np.all(inv.terms_array[:, 2, 0:2] == [n, l]) and np.all(inv.terms_array[:, 0:2, 1] < l)


class InvariantsGenerator:

    @staticmethod
    def delta(l1, l2, l3) -> bool:
        """Delta condition.  Returns True if |l2 - l3| <= l1 <= l2 + l3"""
        return abs(l2 - l3) <= l1 <= l2 + l3

    @staticmethod
    def total_degree_1(index_traits: sph.IndexTraits) -> int:
        """Total number of degree 1 invariants up to n_max"""
        return sum(1 for _ in index_traits.iter_n(0))

    @staticmethod
    def num_degree_2(index_traits: sph.IndexTraits, l: int) -> int:
        n = sum(1 for _ in index_traits.iter_n(l))
        return int(n * (n + 1) / 2)

    @classmethod
    def total_degree_2(cls, index_traits: sph.IndexTraits) -> int:
        l_max = index_traits.l[1]
        return sum(cls.num_degree_2(index_traits, l) for l in utils.inclusive(1, l_max, 1))

    @staticmethod
    def inv_degree_2(n1: int, n2: int, l: int) -> moment_invariants.MomentInvariant:
        """Generate a degree-2 invariant
        """
        builder = moment_invariants.InvariantBuilder(2)
        # This comes from the Clebsch-Gordan coefficient: <l,l,m,-m|0,0> = (-1)^(l-m)/sqrt(2l + 1)
        recip_prefactor = (2 * l + 1)**0.5

        for m in utils.inclusive(-l, l, 1):
            builder.add_term((-1)**(l - m) / recip_prefactor, ((n1, l, m), (n2, l, -m)))

        return builder.build()

    @classmethod
    def inv_degree_3(cls, n1, l1, n2, l2, n3, l3):
        """Generate degree-3 invariant"""
        assert l1 <= n1, f'{l1} > {n1}'
        assert l2 <= n2, f'{l2} > {n2}'
        assert l3 <= n3, f'{l3} > {n3}'
        assert cls.delta(l1, l2, l3)

        builder = moment_invariants.InvariantBuilder(3)

        recip_prefactor = (2 * l1 + 1)**0.5
        for m1 in utils.inclusive(-l1, l1):
            for m2 in utils.inclusive(-l2, l2):
                for m3 in utils.inclusive(-l3, l3):
                    if m1 + m2 + m3 != 0:
                        continue

                    prefactor = (-1)**m1 * complex(cg.CG(l2, m2, l3, m3, l1, -m1).doit()) / recip_prefactor
                    builder.add_term(prefactor, ((n1, l1, m1), (n2, l2, m2), (n3, l3, m3)))

        return builder.build()

    @staticmethod
    def generate_degree_1(index_traits: sph.IndexTraits) -> moment_invariants.MomentInvariants:
        """Generate first degree invariants up to the maximum n"""
        invs = moment_invariants.MomentInvariants()
        for n in index_traits.iter_n(0):
            builder = moment_invariants.InvariantBuilder(1)
            builder.add_term(1, [(n, 0, 0)])
            invs.append(builder.build())

        return invs

    @classmethod
    def generate_degree_2(cls, index_traits: sph.IndexTraits) -> np.ma.masked_array:
        """Generate second degree invariants up to the maximum n, and optionally max l"""
        N = index_traits.N
        L = index_traits.L

        invs_array = np.empty((L, N, N), dtype=object)
        invs_array.fill(None)

        for l in utils.inclusive(1, index_traits.l.max, 1):
            for n1 in index_traits.iter_n(l):
                for n2 in index_traits.iter_n(l, n_spec=(n1, None)):
                    invs_array[l, n1, n2] = cls.inv_degree_2(n1, n2, l)

        return np.ma.masked_array(invs_array, invs_array == None)  # pylint: disable=singleton-comparison

    @classmethod
    def generate_degree_3(cls, index_traits: sph.IndexTraits) -> moment_invariants.MomentInvariants:
        """Generate third degree invariants up to maximum n, and optionally max l"""
        # pylint: disable=too-many-locals
        lmax = index_traits.l[1]
        done = set()

        invs = moment_invariants.MomentInvariants()

        # We need one 3rd degree invariant at l1=l2=l3 that includes three radial terms in order to resolve
        # the sign ambiguity on the unitary matrix from Cholesky decomposition
        for l in utils.inclusive(1, lmax):
            for n1 in index_traits.iter_n(l):
                for n2 in index_traits.iter_n(l, n_spec=(n1, None)):
                    for n3 in index_traits.iter_n(l, n_spec=(n2, None)):
                        if degree_3_is_zero((n1, l), (n2, l), (n3, l)):
                            continue

                        invs.append(cls.inv_degree_3(n1, l, n2, l, n3, l))
                        done.add(((n1, l), (n2, l), (n3, l)))

        for pair_c in index_traits.iter_nl(n_spec=(2, None), l_spec=(2, None)):
            for pair_b in index_traits.iter_nl(l_spec=(None, pair_c[1] - 1)):
                for pair_a in index_traits.iter_nl(l_spec=(None, pair_b[1])):
                    pairs = tuple(sorted([pair_a, pair_b, pair_c], key=lambda p: (p[1], p[0])))
                    # pairs = pair_a, pair_b, pair_c
                    (n1, l1), (n2, l2), (n3, l3) = pairs

                    # Check delta criterion
                    if not cls.delta(l1, l2, l3):
                        continue

                    # Check for permutations that have already been done
                    if pairs in done:
                        continue

                    # Check for zero invariants
                    if degree_3_is_zero(*pairs):
                        continue

                    inv = cls.inv_degree_3(n1, l1, n2, l2, n3, l3)
                    invs.append(inv)
                    done.add(pairs)

        return invs

    @classmethod
    def generate_all(cls, index_traits: sph.IndexTraits) -> InvertibleInvariants:
        """Generate all moments invariants using this scheme up to the max n and l"""
        invs = InvertibleInvariants(
            cls.generate_degree_1(index_traits), cls.generate_degree_2(index_traits),
            cls.generate_degree_3(index_traits), index_traits
        )

        return invs


def generate_all(index_traits: sph.IndexTraits) -> InvertibleInvariants:
    return InvariantsGenerator.generate_all(index_traits)


def degree_3_is_zero(pair1: Tuple, pair2: Tuple, pair3: Tuple) -> bool:
    """Check if a set a degree-3 invariant is identically zero"""
    if mathutil.odd(pair1[1]) and pair1 == pair2 == pair3:
        # All same
        return True

    counts = collections.defaultdict(int)
    counts[pair1] += 1
    counts[pair2] += 1
    counts[pair3] += 1

    if mathutil.odd(pair3[1]) and pair1 == pair2 and pair1 != pair3:
        return True
    if mathutil.odd(pair2[1]) and pair1 == pair3 and pair1 != pair2:
        return True
    if mathutil.odd(pair1[1]) and pair2 == pair3 and pair2 != pair1:
        return True

    return False



# -*- coding: utf-8 -*-
import abc

from milad import atomic
from milad import functions

__all__ = ('Descriptor',)


class Descriptor(functions.Function):
    input_type = atomic.AtomsCollection

    @property
    @abc.abstractmethod
    def fingerprint_len(self) -> int:
        """Get the length of the fingerprint vector that will be returned"""

    @property
    @abc.abstractmethod
    def cutoff(self) -> float:
        """Get the descriptor cutoff radius"""

    @abc.abstractmethod
    def evaluate(self, atoms: atomic.AtomsCollection, *, get_jacobian=False):  # pylint: disable=arguments-differ
        """Custom evaluate for descriptors.  These always take an AtomsCollection"""
